# Copyright (c) 2021, Victor Martinez <VictorMartinezRubio@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
author: Victor Martinez (@v1v)  <VictorMartinezRubio@gmail.com>
name: opentelemetry
type: notification
short_description: Create distributed traces with OpenTelemetry
version_added: 3.7.0
description:
  - This callback creates distributed traces for each Ansible task with OpenTelemetry.
  - You can configure the OpenTelemetry exporter and SDK with environment variables.
  - See U(https://opentelemetry-python.readthedocs.io/en/latest/exporter/otlp/otlp.html).
  - See
    U(https://opentelemetry-python.readthedocs.io/en/latest/sdk/environment_variables.html#opentelemetry-sdk-environment-variables).
options:
  hide_task_arguments:
    default: false
    type: bool
    description:
      - Hide the arguments for a task.
    env:
      - name: ANSIBLE_OPENTELEMETRY_HIDE_TASK_ARGUMENTS
    ini:
      - section: callback_opentelemetry
        key: hide_task_arguments
        version_added: 5.3.0
  enable_from_environment:
    type: str
    description:
      - Whether to enable this callback only if the given environment variable exists and it is set to V(true).
      - This is handy when you use Configuration as Code and want to send distributed traces if running in the CI rather when
        running Ansible locally.
      - For such, it evaluates the given O(enable_from_environment) value as environment variable and if set to V(true) this
        plugin is enabled.
    env:
      - name: ANSIBLE_OPENTELEMETRY_ENABLE_FROM_ENVIRONMENT
    ini:
      - section: callback_opentelemetry
        key: enable_from_environment
        version_added: 5.3.0
    version_added: 3.8.0
  otel_service_name:
    default: ansible
    type: str
    description:
      - The service name resource attribute.
    env:
      - name: OTEL_SERVICE_NAME
    ini:
      - section: callback_opentelemetry
        key: otel_service_name
        version_added: 5.3.0
  traceparent:
    default: None
    type: str
    description:
      - The L(W3C Trace Context header traceparent,https://www.w3.org/TR/trace-context-1/#traceparent-header).
    env:
      - name: TRACEPARENT
  disable_logs:
    default: false
    type: bool
    description:
      - Disable sending logs.
    env:
      - name: ANSIBLE_OPENTELEMETRY_DISABLE_LOGS
    ini:
      - section: callback_opentelemetry
        key: disable_logs
    version_added: 5.8.0
  disable_attributes_in_logs:
    default: false
    type: bool
    description:
      - Disable populating span attributes to the logs.
    env:
      - name: ANSIBLE_OPENTELEMETRY_DISABLE_ATTRIBUTES_IN_LOGS
    ini:
      - section: callback_opentelemetry
        key: disable_attributes_in_logs
    version_added: 7.1.0
  store_spans_in_file:
    type: str
    description:
      - It stores the exported spans in the given file.
    env:
      - name: ANSIBLE_OPENTELEMETRY_STORE_SPANS_IN_FILE
    ini:
      - section: callback_opentelemetry
        key: store_spans_in_file
    version_added: 9.0.0
  otel_exporter_otlp_traces_protocol:
    type: str
    description:
      - E(OTEL_EXPORTER_OTLP_TRACES_PROTOCOL) represents the transport protocol for spans.
      - See
        U(https://opentelemetry-python.readthedocs.io/en/latest/sdk/environment_variables.html#envvar-OTEL_EXPORTER_OTLP_TRACES_PROTOCOL).
    default: grpc
    choices:
      - grpc
      - http/protobuf
    env:
      - name: OTEL_EXPORTER_OTLP_TRACES_PROTOCOL
    ini:
      - section: callback_opentelemetry
        key: otel_exporter_otlp_traces_protocol
    version_added: 9.0.0
requirements:
  - opentelemetry-api (Python library)
  - opentelemetry-exporter-otlp (Python library)
  - opentelemetry-sdk (Python library)
"""


EXAMPLES = r"""
examples: |-
  Enable the plugin in ansible.cfg:
    [defaults]
    callbacks_enabled = community.general.opentelemetry
    [callback_opentelemetry]
    enable_from_environment = ANSIBLE_OPENTELEMETRY_ENABLED

  Set the environment variable:
    export OTEL_EXPORTER_OTLP_ENDPOINT=<your endpoint (OTLP/HTTP)>
    export OTEL_EXPORTER_OTLP_HEADERS="authorization=Bearer your_otel_token"
    export OTEL_SERVICE_NAME=your_service_name
    export ANSIBLE_OPENTELEMETRY_ENABLED=true
"""

import getpass
import json
import os
import socket
import uuid
from collections import OrderedDict
from os.path import basename
from time import time_ns
from urllib.parse import urlparse

from ansible.errors import AnsibleError
from ansible.module_utils.ansible_release import __version__ as ansible_version
from ansible.plugins.callback import CallbackBase

OTEL_LIBRARY_IMPORT_ERROR: ImportError | None
try:
    from opentelemetry import trace
    from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter as GRPCOTLPSpanExporter
    from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter as HTTPOTLPSpanExporter
    from opentelemetry.sdk.resources import SERVICE_NAME, Resource
    from opentelemetry.sdk.trace import TracerProvider
    from opentelemetry.sdk.trace.export import BatchSpanProcessor, SimpleSpanProcessor
    from opentelemetry.sdk.trace.export.in_memory_span_exporter import InMemorySpanExporter
    from opentelemetry.trace import SpanKind
    from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator
    from opentelemetry.trace.status import Status, StatusCode
except ImportError as imp_exc:
    OTEL_LIBRARY_IMPORT_ERROR = imp_exc
else:
    OTEL_LIBRARY_IMPORT_ERROR = None


class TaskData:
    """
    Data about an individual task.
    """

    def __init__(self, uuid, name, path, play, action, args):
        self.uuid = uuid
        self.name = name
        self.path = path
        self.play = play
        self.host_data = OrderedDict()
        self.start = time_ns()
        self.action = action
        self.args = args
        self.dump = None

    def add_host(self, host):
        if host.uuid in self.host_data:
            if host.status == "included":
                # concatenate task include output from multiple items
                host.result = f"{self.host_data[host.uuid].result}\n{host.result}"
            else:
                self.host_data[host.uuid].update(host.status, host.result)
                return

        self.host_data[host.uuid] = host


class HostData:
    """
    Data about an individual host.
    """

    def __init__(self, uuid, name, status, result=None):
        self.uuid = uuid
        self.name = name
        self.status = status
        self.result = result
        self.finish = None
        self.start = time_ns()

    def update(self, status, result):
        self.status = status
        self.result = result
        self.finish = time_ns()


class OpenTelemetrySource:
    def __init__(self, display):
        self.ansible_playbook = ""
        self.session = str(uuid.uuid4())
        self.host = socket.gethostname()
        try:
            self.ip_address = socket.gethostbyname(socket.gethostname())
        except Exception:
            self.ip_address = None
        self.user = getpass.getuser()

        self._display = display

    def traceparent_context(self, traceparent):
        carrier = dict()
        carrier["traceparent"] = traceparent
        return TraceContextTextMapPropagator().extract(carrier=carrier)

    def start_task(self, tasks_data, hide_task_arguments, play_name, task, host):
        """record the start of a task for one or more hosts"""

        uuid = task._uuid

        if uuid in tasks_data:
            tasks_data[uuid].add_host(HostData(host._uuid, host.name, "started"))
            return

        name = task.get_name().strip()
        path = task.get_path()
        action = task.action
        args = None

        if not task.no_log and not hide_task_arguments:
            args = task.args

        tasks_data[uuid] = TaskData(uuid, name, path, play_name, action, args)
        tasks_data[uuid].add_host(HostData(host._uuid, host.name, "started"))

    def finish_task(self, tasks_data, status, result, dump):
        """record the results of a task for a single host"""

        task_uuid = result._task._uuid

        if hasattr(result, "_host") and result._host is not None:
            host_uuid = result._host._uuid
            host_name = result._host.name
        else:
            host_uuid = "include"
            host_name = "include"

        task = tasks_data[task_uuid]

        task.dump = dump
        task.add_host(HostData(host_uuid, host_name, status, result))

    def generate_distributed_traces(
        self,
        otel_service_name,
        ansible_playbook,
        tasks_data,
        status,
        traceparent,
        disable_logs,
        disable_attributes_in_logs,
        otel_exporter_otlp_traces_protocol,
        store_spans_in_file,
    ):
        """generate distributed traces from the collected TaskData and HostData"""

        tasks = []
        parent_start_time = None
        for task in tasks_data.values():
            if parent_start_time is None:
                parent_start_time = task.start
            tasks.append(task)

        trace.set_tracer_provider(TracerProvider(resource=Resource.create({SERVICE_NAME: otel_service_name})))

        otel_exporter = None
        if store_spans_in_file:
            otel_exporter = InMemorySpanExporter()
            processor = SimpleSpanProcessor(otel_exporter)
        else:
            if otel_exporter_otlp_traces_protocol == "grpc":
                otel_exporter = GRPCOTLPSpanExporter()
            else:
                otel_exporter = HTTPOTLPSpanExporter()
            processor = BatchSpanProcessor(otel_exporter)

        trace.get_tracer_provider().add_span_processor(processor)

        tracer = trace.get_tracer(__name__)

        with tracer.start_as_current_span(
            ansible_playbook,
            context=self.traceparent_context(traceparent),
            start_time=parent_start_time,
            kind=SpanKind.SERVER,
        ) as parent:
            parent.set_status(status)
            # Populate trace metadata attributes
            parent.set_attribute("ansible.version", ansible_version)
            parent.set_attribute("ansible.session", self.session)
            parent.set_attribute("ansible.host.name", self.host)
            if self.ip_address is not None:
                parent.set_attribute("ansible.host.ip", self.ip_address)
            parent.set_attribute("ansible.host.user", self.user)
            for task in tasks:
                for host_data in task.host_data.values():
                    start = host_data.start or task.start
                    with tracer.start_as_current_span(task.name, start_time=start, end_on_exit=False) as span:
                        self.update_span_data(task, host_data, span, disable_logs, disable_attributes_in_logs)

        return otel_exporter

    def update_span_data(self, task_data, host_data, span, disable_logs, disable_attributes_in_logs):
        """update the span with the given TaskData and HostData"""

        name = f"[{host_data.name}] {task_data.play}: {task_data.name}"

        message = "success"
        res = {}
        rc = 0
        status = Status(status_code=StatusCode.OK)
        if host_data.status != "included":
            # Support loops
            enriched_error_message = None
            if host_data.result and "results" in host_data.result._result:
                if host_data.status == "failed":
                    message = self.get_error_message_from_results(host_data.result._result["results"], task_data.action)
                    enriched_error_message = self.enrich_error_message_from_results(
                        host_data.result._result["results"], task_data.action
                    )
            elif host_data.result:
                res = host_data.result._result
                rc = res.get("rc", 0)
                if host_data.status == "failed":
                    message = self.get_error_message(res)
                    enriched_error_message = self.enrich_error_message(res)

            if host_data.status == "failed":
                status = Status(status_code=StatusCode.ERROR, description=message)
                # Record an exception with the task message
                span.record_exception(BaseException(enriched_error_message))
            elif host_data.status == "skipped":
                message = res["skip_reason"] if "skip_reason" in res else "skipped"
                status = Status(status_code=StatusCode.UNSET)
            elif host_data.status == "ignored":
                status = Status(status_code=StatusCode.UNSET)

        span.set_status(status)

        # Create the span and log attributes
        attributes = {
            "ansible.task.module": task_data.action,
            "ansible.task.message": message,
            "ansible.task.name": name,
            "ansible.task.result": rc,
            "ansible.task.host.name": host_data.name,
            "ansible.task.host.status": host_data.status,
        }
        if isinstance(task_data.args, dict) and "gather_facts" not in task_data.action:
            names = tuple(self.transform_ansible_unicode_to_str(k) for k in task_data.args.keys())
            values = tuple(self.transform_ansible_unicode_to_str(k) for k in task_data.args.values())
            attributes[("ansible.task.args.name")] = names
            attributes[("ansible.task.args.value")] = values

        self.set_span_attributes(span, attributes)

        # This will allow to enrich the service map
        self.add_attributes_for_service_map_if_possible(span, task_data)
        # Send logs
        if not disable_logs:
            # This will avoid populating span attributes to the logs
            span.add_event(task_data.dump, attributes={} if disable_attributes_in_logs else attributes)
        # Close span always
        span.end(end_time=host_data.finish)

    def set_span_attributes(self, span, attributes):
        """update the span attributes with the given attributes if not None"""

        if span is None and self._display is not None:
            self._display.warning("span object is None. Please double check if that is expected.")
        else:
            if attributes is not None:
                span.set_attributes(attributes)

    def add_attributes_for_service_map_if_possible(self, span, task_data):
        """Update the span attributes with the service that the task interacted with, if possible."""

        redacted_url = self.parse_and_redact_url_if_possible(task_data.args)
        if redacted_url:
            span.set_attribute("http.url", redacted_url.geturl())

    @staticmethod
    def parse_and_redact_url_if_possible(args):
        """Parse and redact the url, if possible."""

        try:
            parsed_url = urlparse(OpenTelemetrySource.url_from_args(args))
        except ValueError:
            return None

        if OpenTelemetrySource.is_valid_url(parsed_url):
            return OpenTelemetrySource.redact_user_password(parsed_url)
        return None

    @staticmethod
    def url_from_args(args):
        # the order matters
        url_args = (
            "url",
            "api_url",
            "baseurl",
            "repo",
            "server_url",
            "chart_repo_url",
            "registry_url",
            "endpoint",
            "uri",
            "updates_url",
        )
        for arg in url_args:
            if args is not None and args.get(arg):
                return args.get(arg)
        return ""

    @staticmethod
    def redact_user_password(url):
        return url._replace(netloc=url.hostname) if url.password else url

    @staticmethod
    def is_valid_url(url):
        if all([url.scheme, url.netloc, url.hostname]):
            return "{{" not in url.hostname
        return False

    @staticmethod
    def transform_ansible_unicode_to_str(value):
        parsed_url = urlparse(str(value))
        if OpenTelemetrySource.is_valid_url(parsed_url):
            return OpenTelemetrySource.redact_user_password(parsed_url).geturl()
        return str(value)

    @staticmethod
    def get_error_message(result):
        if result.get("exception") is not None:
            return OpenTelemetrySource._last_line(result["exception"])
        return result.get("msg", "failed")

    @staticmethod
    def get_error_message_from_results(results, action):
        for result in results:
            if result.get("failed", False):
                return f"{action}({result.get('item', 'none')}) - {OpenTelemetrySource.get_error_message(result)}"

    @staticmethod
    def _last_line(text):
        lines = text.strip().split("\n")
        return lines[-1]

    @staticmethod
    def enrich_error_message(result):
        message = result.get("msg", "failed")
        exception = result.get("exception")
        stderr = result.get("stderr")
        return f'message: "{message}"\nexception: "{exception}"\nstderr: "{stderr}"'

    @staticmethod
    def enrich_error_message_from_results(results, action):
        message = ""
        for result in results:
            if result.get("failed", False):
                message = f"{action}({result.get('item', 'none')}) - {OpenTelemetrySource.enrich_error_message(result)}\n{message}"
        return message


class CallbackModule(CallbackBase):
    """
    This callback creates distributed traces.
    """

    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = "notification"
    CALLBACK_NAME = "community.general.opentelemetry"
    CALLBACK_NEEDS_ENABLED = True

    def __init__(self, display=None):
        super().__init__(display=display)
        self.hide_task_arguments = None
        self.disable_attributes_in_logs = None
        self.disable_logs = None
        self.otel_service_name = None
        self.ansible_playbook = None
        self.play_name = None
        self.tasks_data = None
        self.errors = 0
        self.disabled = False
        self.traceparent = False
        self.store_spans_in_file = False
        self.otel_exporter_otlp_traces_protocol = None

        if OTEL_LIBRARY_IMPORT_ERROR:
            raise AnsibleError(
                "The `opentelemetry-api`, `opentelemetry-exporter-otlp` or `opentelemetry-sdk` must be installed to use this plugin"
            ) from OTEL_LIBRARY_IMPORT_ERROR

        self.tasks_data = OrderedDict()

        self.opentelemetry = OpenTelemetrySource(display=self._display)

    def set_options(self, task_keys=None, var_options=None, direct=None):
        super().set_options(task_keys=task_keys, var_options=var_options, direct=direct)

        environment_variable = self.get_option("enable_from_environment")
        if environment_variable is not None and os.environ.get(environment_variable, "false").lower() != "true":
            self.disabled = True
            self._display.warning(
                f"The `enable_from_environment` option has been set and {environment_variable} is not enabled. Disabling the `opentelemetry` callback plugin."
            )

        self.hide_task_arguments = self.get_option("hide_task_arguments")

        self.disable_attributes_in_logs = self.get_option("disable_attributes_in_logs")

        self.disable_logs = self.get_option("disable_logs")

        self.store_spans_in_file = self.get_option("store_spans_in_file")

        self.otel_service_name = self.get_option("otel_service_name")

        if not self.otel_service_name:
            self.otel_service_name = "ansible"

        # See https://github.com/open-telemetry/opentelemetry-specification/issues/740
        self.traceparent = self.get_option("traceparent")

        self.otel_exporter_otlp_traces_protocol = self.get_option("otel_exporter_otlp_traces_protocol")

    def dump_results(self, task, result):
        """dump the results if disable_logs is not enabled"""
        if self.disable_logs:
            return ""
        # ansible.builtin.uri contains the response in the json field
        save = dict(result._result)

        if "json" in save and task.action in ("ansible.builtin.uri", "ansible.legacy.uri", "uri"):
            save.pop("json")
        # ansible.builtin.slurp contains the response in the content field
        if "content" in save and task.action in ("ansible.builtin.slurp", "ansible.legacy.slurp", "slurp"):
            save.pop("content")
        return self._dump_results(save)

    def v2_playbook_on_start(self, playbook):
        self.ansible_playbook = basename(playbook._file_name)

    def v2_playbook_on_play_start(self, play):
        self.play_name = play.get_name()

    def v2_runner_on_start(self, host, task):
        self.opentelemetry.start_task(self.tasks_data, self.hide_task_arguments, self.play_name, task, host)

    def v2_runner_on_failed(self, result, ignore_errors=False):
        if ignore_errors:
            status = "ignored"
        else:
            status = "failed"
            self.errors += 1

        self.opentelemetry.finish_task(
            self.tasks_data, status, result, self.dump_results(self.tasks_data[result._task._uuid], result)
        )

    def v2_runner_on_ok(self, result):
        self.opentelemetry.finish_task(
            self.tasks_data, "ok", result, self.dump_results(self.tasks_data[result._task._uuid], result)
        )

    def v2_runner_on_skipped(self, result):
        self.opentelemetry.finish_task(
            self.tasks_data, "skipped", result, self.dump_results(self.tasks_data[result._task._uuid], result)
        )

    def v2_runner_on_unreachable(self, result):
        self.errors += 1
        self.opentelemetry.finish_task(
            self.tasks_data, "failed", result, self.dump_results(self.tasks_data[result._task._uuid], result)
        )

    def v2_playbook_on_include(self, included_file):
        self.opentelemetry.finish_task(self.tasks_data, "included", included_file, "")

    def v2_playbook_on_stats(self, stats):
        if self.errors == 0:
            status = Status(status_code=StatusCode.OK)
        else:
            status = Status(status_code=StatusCode.ERROR)
        otel_exporter = self.opentelemetry.generate_distributed_traces(
            self.otel_service_name,
            self.ansible_playbook,
            self.tasks_data,
            status,
            self.traceparent,
            self.disable_logs,
            self.disable_attributes_in_logs,
            self.otel_exporter_otlp_traces_protocol,
            self.store_spans_in_file,
        )

        if self.store_spans_in_file:
            spans = [json.loads(span.to_json()) for span in otel_exporter.get_finished_spans()]
            with open(self.store_spans_in_file, "w", encoding="utf-8") as output:
                json.dump({"spans": spans}, output, indent=4)

    def v2_runner_on_async_failed(self, result, **kwargs):
        self.errors += 1



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: ucs_system_qos
short_description: Configures system QoS settings
description:
   -  Configures system QoS settings
extends_documentation_fragment: cisco.ucs.ucs
options:
    priority:
        description: Priority to configure
        choices: ["best-effort", "bronze", "fc", "gold","platinum", "silver"]
        required: true
        type: str
    admin_state:
        description: Admin state of QoS Policy
        choices: ['disabled', 'enabled']
        default: enabled
        type: str
    cos:
        description: CoS setting. Choices are any or 0-6.
        required: true
        type: str
    weight:
        description: CoS profile weight. Choices are best-effort, none or 0-10.
        required: true
        type: str
    mtu:
        description: MTU size. Choices are fc, normal or 0-4294967295.
        default: normal
        type: str
    multicast_optimize:
        description: Set multicast optimization options
        choices: ['false', 'no', 'true', 'yes']
        default: 'no'
        type: str
    drop:
        description: Set multicast optimization options
        default: 'drop'
        choices: ['drop', 'no-drop']
        type: str
requirements: ['ucsmsdk']
author: "Brett Johnson (@sdbrett)"
'''

EXAMPLES = '''
- name:
  cisco.ucs.ucs_system_qos:
    priority: platinum
    admin_state: enabled
    multicast_optimize: no
    cos: '5'
    weight: '10'
    mtu: '9216'
    hostname: 192.168.99.100
    username: admin
    password: password
'''

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.ucs.plugins.module_utils.ucs import UCSModule, ucs_argument_spec


# TODO Add ranges for cos, weight and mtu
def main():
    argument_spec = ucs_argument_spec.copy()
    argument_spec.update(
        priority=dict(required=True, type='str', choices=["best-effort", "bronze", "fc", "gold", "platinum", "silver"]),
        cos=dict(required=True, type='str'),
        weight=dict(required=True, type='str'),
        admin_state=dict(required=False, type='str', default='enabled', choices=['disabled', 'enabled']),
        drop=dict(required=False, type='str', default='drop', choices=['drop', 'no-drop']),
        mtu=dict(required=False, type='str', default='normal'),
        multicast_optimize=dict(required=False, type='str', default='no', choices=['false', 'no', 'true', 'yes']),
    )

    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
    )
    ucs = UCSModule(module)

    err = False

    changed = False
    try:
        dn = "fabric/lan/classes/class-" + module.params['priority']
        mo = ucs.login_handle.query_dn(dn)
        # check top-level mo props
        if module.params['priority'] == 'best-effort':
            kwargs = dict(weight=module.params['weight'])
            kwargs['mtu'] = module.params['mtu']
            kwargs['multicast_optimize'] = module.params['multicast_optimize']
            if not mo.check_prop_match(**kwargs):
                if not module.check_mode:
                    mo.weight = module.params['weight']
                    mo.mtu = module.params['mtu']
                    mo.multicast_optimize = module.params['multicast_optimize']
                    ucs.login_handle.add_mo(mo, True)
                    ucs.login_handle.commit()
                changed = True
        elif module.params['priority'] == 'fc':
            kwargs = dict(weight=module.params['weight'])
            kwargs['cos'] = module.params['cos']
            if not mo.check_prop_match(**kwargs):
                if not module.check_mode:
                    mo.weight = module.params['weight']
                    mo.cos = module.params['cos']
                    ucs.login_handle.add_mo(mo, True)
                    ucs.login_handle.commit()
                changed = True

        else:
            kwargs = dict(weight=module.params['weight'])
            kwargs['priority'] = module.params['priority']
            kwargs['mtu'] = module.params['mtu']
            kwargs['cos'] = module.params['cos']
            kwargs['drop'] = module.params['drop']
            kwargs['admin_state'] = module.params['admin_state']
            kwargs['multicast_optimize'] = module.params['multicast_optimize']
        if not mo.check_prop_match(**kwargs):
            if not module.check_mode:
                mo.weight = module.params['weight']
                mo.mtu = module.params['mtu']
                mo.cos = module.params['cos']
                mo.drop = module.params['drop']
                mo.admin_state = module.params['admin_state']
                mo.multicast_optimize = module.params['multicast_optimize']

                ucs.login_handle.add_mo(mo, True)
                ucs.login_handle.commit()
            changed = True

    except Exception as e:
        err = True
        ucs.result['msg'] = "setup error: %s " % str(e)

    ucs.result['changed'] = changed
    if err:
        module.fail_json(**ucs.result)
    module.exit_json(**ucs.result)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The Bgp_address_family parser templates file. This contains
a list of parser definitions and associated functions that
facilitates both facts gathering and native command generation for
the given network resource.
"""

import re

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)


def _tmplt_aggregate_address(aggaddr):
    cmd = "aggregate-address {prefix}"

    if aggaddr.get("advertise_map"):
        cmd += " advertise-map {advertise_map}"
    if aggaddr.get("as_set"):
        cmd += " as-set"
    if aggaddr.get("attribute_map"):
        cmd += " attribute-map {attribute_map}"
    if aggaddr.get("summary_only"):
        cmd += " summary-only"
    if aggaddr.get("suppress_map"):
        cmd += " suppress-map {suppress_map}"

    return cmd.format(**aggaddr)


def _tmplt_dampening(proc):
    damp = proc.get("dampening", {})
    cmd = "dampening"

    if damp.get("set") is False:
        return "no {0}".format(cmd)
    if damp.get("route_map"):
        cmd += " route-map {route_map}".format(**damp)
    for x in (
        "decay_half_life",
        "start_reuse_route",
        "start_suppress_route",
        "max_suppress_time",
    ):
        if x in damp:
            cmd += " {0}".format(damp[x])
    return cmd


def _tmplt_redistribute(redis):
    command = "redistribute {protocol}".format(**redis)
    if redis.get("id"):
        command += " {id}".format(**redis)
    command += " route-map {route_map}".format(**redis)
    return command


class Bgp_address_familyTemplate(NetworkTemplate):
    def __init__(self, lines=None):
        super(Bgp_address_familyTemplate, self).__init__(lines=lines, tmplt=self)

    # fmt: off
    PARSERS = [
        {
            "name": "as_number",
            "getval": re.compile(
                r"""
                ^router\sbgp\s(?P<as_number>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "router bgp {{ as_number }}",
            "result": {
                "as_number": "{{ as_number }}",
            },
            "shared": True,
        },
        {
            "name": "address_family",
            "getval": re.compile(
                r"""
                (\s+vrf\s(?P<vrf>\S+))?
                (\s+neighbor\s(?P<nbr>\S+))?
                \s+address-family
                \s(?P<afi>\S+)
                (\s(?P<safi>\S+))?
                $""",
                re.VERBOSE,
            ),
            "setval": "address-family {{ afi }}{{ (' ' + safi) if safi is defined else ''}}",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "vrf": "{{ vrf }}",
                        "afi": "{{ afi }}",
                        "safi": "{{ safi }}",
                    },
                },
            },
            "shared": True,
        },
        {
            "name": "additional_paths.install_backup",
            "getval": re.compile(
                r"""
                \s+additional-paths
                \sinstall\s(?P<backup>backup)
                $""",
                re.VERBOSE,
            ),
            "setval": "additional-paths install backup",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "additional_paths": {
                            "install_backup": "{{ not not backup }}",
                        },
                    },
                },
            },
        },
        {
            "name": "additional_paths.receive",
            "getval": re.compile(
                r"""
                \s+additional-paths
                \s(?P<receive>receive)
                $""",
                re.VERBOSE,
            ),
            "setval": "additional-paths receive",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "additional_paths": {
                            "receive": "{{ not not receive }}",
                        },
                    },
                },
            },
        },
        {
            "name": "additional_paths.selection.route_map",
            "getval": re.compile(
                r"""
                \s+additional-paths
                \sselection\sroute-map
                \s(?P<route_map>\S+)
                $""",
                re.VERBOSE,
            ),
            "setval": "additional-paths selection route-map {{ additional_paths.selection.route_map }}",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "additional_paths": {
                            "selection": {
                                "route_map": "{{ route_map }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "additional_paths.send",
            "getval": re.compile(
                r"""
                \s+additional-paths
                \s(?P<send>send)
                $""",
                re.VERBOSE,
            ),
            "setval": "additional-paths send",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "additional_paths": {
                            "send": "{{ not not send }}",
                        },
                    },
                },
            },
        },
        {
            "name": "advertise_l2vpn_evpn",
            "getval": re.compile(
                r"""
                \s+(?P<advertise_l2vpn_evpn>advertise\sl2vpn\sevpn)
                $""",
                re.VERBOSE,
            ),
            "setval": "advertise l2vpn evpn",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "advertise_l2vpn_evpn": "{{ not not advertise_l2vpn_evpn }}",
                    },
                },
            },
        },
        {
            "name": "advertise_pip",
            "getval": re.compile(
                r"""
                \s+(?P<advertise_pip>advertise-pip)
                $""",
                re.VERBOSE,
            ),
            "setval": "advertise-pip",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "advertise_pip": "{{ not not advertise_pip }}",
                    },
                },
            },
        },
        {
            "name": "advertise_system_mac",
            "getval": re.compile(
                r"""
                \s+(?P<advertise_system_mac>advertise-system-mac)
                $""",
                re.VERBOSE,
            ),
            "setval": "advertise-system-mac",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "advertise_system_mac": "{{ not not advertise_system_mac }}",
                    },
                },
            },
        },
        {
            "name": "allow_vni_in_ethertag",
            "getval": re.compile(
                r"""
                \s+(?P<allow_vni_in_ethertag>allow-vni-in-ethertag)
                $""",
                re.VERBOSE,
            ),
            "setval": "allow-vni-in-ethertag",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "allow_vni_in_ethertag": "{{ not not allow_vni_in_ethertag }}",
                    },
                },
            },
        },
        {
            "name": "aggregate_address",
            "getval": re.compile(
                r"""
                \s+aggregate-address
                \s(?P<prefix>\S+)
                (\s(?P<as_set>as-set))?
                (\s(?P<summary_only>summary-only))?
                (\sadvertise-map\s(?P<advertise_map>\S+))?
                (\sattribute-map\s(?P<attribute_map>\S+))?
                (\ssuppress-map\s(?P<suppress_map>\S+))?
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_aggregate_address,
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "aggregate_address": [
                            {
                                "prefix": "{{ prefix }}",
                                "as_set": "{{ True if as_set is defined else None }}",
                                "summary_only": "{{ True if summary_only is defined else None }}",
                                "advertise_map": "{{ advertise_map }}",
                                "attribute_map": "{{ attribute_map }}",
                                "suppress_map": "{{ suppress_map }}",
                            },
                        ],
                    },
                },
            },
        },
        {
            "name": "client_to_client.no_reflection",
            "getval": re.compile(
                r"""
                \s+no\sclient-to-client
                \s(?P<reflection>reflection)
                $""",
                re.VERBOSE,
            ),
            "setval": "client-to-client reflection",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "client_to_client": {
                            "no_reflection": "{{ not not reflection }}",
                        },
                    },
                },
            },
        },
        {
            "name": "dampen_igp_metric",
            "getval": re.compile(
                r"""
                \s+dampen-igp-metric
                \s(?P<dampen_igp_metric>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": "dampen-igp-metric {{ dampen_igp_metric }}",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "dampen_igp_metric": "{{ dampen_igp_metric }}",
                    },
                },
            },
        },
        {
            "name": "dampening",
            "getval": re.compile(
                r"""
                \s+(?P<dampening>dampening)
                (\s(?P<decay_half_life>\d+))?
                (\s(?P<start_reuse_route>\d+))?
                (\s(?P<start_suppress_route>\d+))?
                (\s(?P<max_suppress_time>\d+))?
                (\sroute-map\s(?P<route_map>\S+))?
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_dampening,
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "dampening": {
                            "set": "{{ True if dampening is defined"
                                   " and ((not decay_half_life|d(False),"
                                   " not start_reuse_route|d(False), "
                                   " not start_suppress_route|d(False), not max_suppress_time|d(False), not route_map|d(""))|all) }}",
                            "decay_half_life": "{{ decay_half_life }}",
                            "start_reuse_route": "{{ start_reuse_route }}",
                            "start_suppress_route": "{{ start_suppress_route }}",
                            "max_suppress_time": "{{ max_suppress_time }}",
                            "route_map": "{{ route_map }}",
                        },
                    },
                },
            },
        },
        {
            "name": "default_information.originate",
            "getval": re.compile(
                r"""
                \s+default-information
                \s(?P<originate>originate)
                $""",
                re.VERBOSE,
            ),
            "setval": "default-information originate",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "default_information": {
                            "originate": "{{ not not originate }}",
                        },
                    },
                },
            },
        },
        {
            "name": "default_metric",
            "getval": re.compile(
                r"""
                \s+default-metric
                \s(?P<default_metric>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": "default-metric {{ default_metric }}",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "default_metric": "{{ default_metric }}",
                    },
                },
            },
        },
        {
            "name": "distance",
            "getval": re.compile(
                r"""
                \s+distance
                \s(?P<ebgp_routes>\d+)
                \s(?P<ibgp_routes>\d+)
                \s(?P<local_routes>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": "distance {{ distance.ebgp_routes }} {{ distance.ibgp_routes }} {{ distance.local_routes }}",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "distance": {
                            "ebgp_routes": "{{ ebgp_routes }}",
                            "ibgp_routes": "{{ ibgp_routes }}",
                            "local_routes": "{{ local_routes }}",
                        },
                    },
                },
            },
        },
        {
            "name": "export_gateway_ip",
            "getval": re.compile(
                r"""
                \s+(?P<export_gateway_ip>export-gateway-ip)
                $""",
                re.VERBOSE,
            ),
            "setval": "export-gateway-ip",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "export_gateway_ip": "{{ not not export_gateway_ip }}",
                    },
                },
            },
        },
        {
            "name": "inject_map",
            "getval": re.compile(
                r"""
                \s+inject-map
                \s(?P<route_map>\S+)
                \sexist-map\s(?P<exist_map>\S+)
                (\s(?P<copy_attributes>copy-attributes))?
                $""",
                re.VERBOSE,
            ),
            "setval": "inject-map {{ route_map }} exist-map {{ exist_map }}{{ ' copy-attributes' if copy_attributes|d(False) else '' }}",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "inject_map": [
                            {
                                "route_map": "{{ route_map }}",
                                "exist_map": "{{ exist_map }}",
                                "copy_attributes": "{{ not not copy_attributes }}",
                            },
                        ],
                    },
                },
            },
        },
        {
            "name": "maximum_paths.parallel_paths",
            "getval": re.compile(
                r"""
                \s+maximum-paths
                \s(?P<parallel_paths>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": "maximum-paths {{ maximum_paths.parallel_paths }}",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "maximum_paths": {
                            "parallel_paths": "{{ parallel_paths }}",
                        },
                    },
                },
            },
        },
        {
            "name": "maximum_paths.ibgp.parallel_paths",
            "getval": re.compile(
                r"""
                \s+maximum-paths
                \sibgp\s(?P<parallel_paths>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": "maximum-paths ibgp {{ maximum_paths.ibgp.parallel_paths }}",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "maximum_paths": {
                            "ibgp": {
                                "parallel_paths": "{{ parallel_paths }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "maximum_paths.eibgp.parallel_paths",
            "getval": re.compile(
                r"""
                \s+maximum-paths
                \seibgp\s(?P<parallel_paths>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": "maximum-paths eibgp {{ maximum_paths.eibgp.parallel_paths }}",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "maximum_paths": {
                            "eibgp": {
                                "parallel_paths": "{{ parallel_paths }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "maximum_paths.local.parallel_paths",
            "getval": re.compile(
                r"""
                \s+maximum-paths
                \slocal\s(?P<parallel_paths>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": "maximum-paths local {{ maximum_paths.local.parallel_paths }}",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "maximum_paths": {
                            "local": {
                                "parallel_paths": "{{ parallel_paths }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "maximum_paths.mixed.parallel_paths",
            "getval": re.compile(
                r"""
                \s+maximum-paths
                \smixed\s(?P<parallel_paths>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": "maximum-paths mixed {{ maximum_paths.mixed.parallel_paths }}",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "maximum_paths": {
                            "mixed": {
                                "parallel_paths": "{{ parallel_paths }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "networks",
            "getval": re.compile(
                r"""
                \s+network
                \s(?P<prefix>\S+)
                (\sroute-map\s(?P<route_map>\S+))?
                $""",
                re.VERBOSE,
            ),
            "setval": "network {{ prefix }}{{ (' route-map ' + route_map) if route_map is defined else '' }}",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "networks": [
                            {
                                "prefix": "{{ prefix }}",
                                "route_map": "{{ route_map }}",
                            },
                        ],
                    },
                },
            },
        },
        {
            "name": "nexthop.route_map",
            "getval": re.compile(
                r"""
                \s+nexthop
                \sroute-map\s(?P<route_map>\S+)
                $""",
                re.VERBOSE,
            ),
            "setval": "nexthop route-map {{ nexthop.route_map }}",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "nexthop": {
                            "route_map": "{{ route_map }}",
                        },
                    },
                },
            },
        },
        {
            "name": "nexthop.trigger_delay",
            "getval": re.compile(
                r"""
                \s+nexthop
                \strigger-delay
                \scritical\s(?P<critical_delay>\d+)
                \snon-critical\s(?P<non_critical_delay>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": "nexthop trigger-delay critical {{ nexthop.trigger_delay.critical_delay }} non-critical {{ nexthop.trigger_delay.non_critical_delay }}",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "nexthop": {
                            "trigger_delay": {
                                "critical_delay": "{{ critical_delay }}",
                                "non_critical_delay": "{{ non_critical_delay }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "redistribute",
            "getval": re.compile(
                r"""
                \s+redistribute
                \s(?P<protocol>\S+)
                (\s(?P<id>\S+))?
                \sroute-map\s(?P<rmap>\S+)
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_redistribute,
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "redistribute": [
                            {
                                "protocol": "{{ protocol }}",
                                "id": "{{ id }}",
                                "route_map": "{{ rmap }}",
                            },
                        ],
                    },
                },
            },
        },
        {
            "name": "retain.route_target.retain_all",
            "getval": re.compile(
                r"""
                \s+retain\sroute-target
                \s(?P<retain_all>all)
                $""",
                re.VERBOSE,
            ),
            "setval": "retain route-target all",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "retain": {
                            "route_target": {
                                "retain_all": "{{ not not retain_all }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "retain.route_target.route_map",
            "getval": re.compile(
                r"""
                \s+retain\sroute-target
                \sroute-map\s(?P<route_map>\S+)
                $""",
                re.VERBOSE,
            ),
            "setval": "retain route-target route-map {{ retain.route_target.route_map }}",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "retain": {
                            "route_target": {
                                "route_map": "{{ route_map }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "suppress_inactive",
            "getval": re.compile(
                r"""
                \s+(?P<suppress_inactive>suppress-inactive)
                $""",
                re.VERBOSE,
            ),
            "setval": "suppress-inactive",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "suppress_inactive": "{{ not not suppress_inactive }}",
                    },
                },
            },
        },
        {
            "name": "table_map",
            "getval": re.compile(
                r"""
                \s+table-map
                \s(?P<name>\S+)
                (\s(?P<filter>filter))?
                $""",
                re.VERBOSE,
            ),
            "setval": "table-map {{ table_map.name }}{{ ' filter' if table_map.filter|d(False) }}",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "table_map": {
                            "name": "{{ name }}",
                            "filter": "{{ not not filter }}",
                        },
                    },
                },
            },
        },
        {
            "name": "timers.bestpath_defer",
            "getval": re.compile(
                r"""
                \s+timers
                \sbestpath-defer\s(?P<defer_time>\d+)
                \smaximum\s(?P<maximum_defer_time>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": "timers bestpath-defer {{ timers.bestpath_defer.defer_time }} maximum {{ timers.bestpath_defer.maximum_defer_time }}",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "timers": {
                            "bestpath_defer": {
                                "defer_time": "{{ defer_time }}",
                                "maximum_defer_time": "{{ maximum_defer_time }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "wait_igp_convergence",
            "getval": re.compile(
                r"""
                \s+(?P<wait_igp_convergence>wait-igp-convergence)
                $""",
                re.VERBOSE,
            ),
            "setval": "wait-igp-convergence",
            "result": {
                "address_family": {
                    '{{ nbr|d("nbr_") + afi + "_" + safi|d() + "_" + vrf|d() }}': {
                        "wait_igp_convergence": "{{ not not wait_igp_convergence }}",
                    },
                },
            },
        },
    ]
    # fmt: on



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: networks_wireless_ssids_traffic_shaping_rules
short_description: Resource module for networks _wireless _ssids _traffic _shaping
  _rules
description:
  - Manage operation update of the resource networks _wireless _ssids _traffic _shaping
    _rules.
  - Update the traffic shaping rules for an SSID on an MR network.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  defaultRulesEnabled:
    description: Whether default traffic shaping rules are enabled (true) or disabled
      (false). There are 4 default rules, which can be seen on your network's traffic
      shaping page. Note that default rules count against the rule limit of 8.
    type: bool
  networkId:
    description: NetworkId path parameter. Network ID.
    type: str
  number:
    description: Number path parameter.
    type: str
  rules:
    description: An array of traffic shaping rules. Rules are applied in the order
      that they are specified in. An empty list (or null) means no rules. Note that
      you are allowed a maximum of 8 rules.
    elements: dict
    suboptions:
      definitions:
        description: A list of objects describing the definitions of your traffic
          shaping rule. At least one definition is required.
        elements: dict
        suboptions:
          type:
            description: The type of definition. Can be one of 'application', 'applicationCategory',
              'host', 'port', 'ipRange' or 'localNet'.
            type: str
          value:
            description: If "type" is 'host', 'port', 'ipRange' or 'localNet', then
              "value" must be a string, matching either a hostname (e.g. "somesite.com"),
              a port (e.g. 8080), or an IP range ("192.1.0.0", "192.1.0.0/16", or
              "10.1.0.0/16 80"). 'localNet' also supports CIDR notation, excluding
              custom ports. If "type" is 'application' or 'applicationCategory', then
              "value" must be an object with the structure { "id" "meraki layer7/..."
              }, where "id" is the application category or application ID (for a list
              of IDs for your network, use the trafficShaping/applicationCategories
              endpoint).
            type: str
        type: list
      dscpTagValue:
        description: The DSCP tag applied by your rule. Null means 'Do not change
          DSCP tag'. For a list of possible tag values, use the trafficShaping/dscpTaggingOptions
          endpoint.
        type: int
      pcpTagValue:
        description: The PCP tag applied by your rule. Can be 0 (lowest priority)
          through 7 (highest priority). Null means 'Do not set PCP tag'.
        type: int
      perClientBandwidthLimits:
        description: An object describing the bandwidth settings for your rule.
        suboptions:
          bandwidthLimits:
            description: The bandwidth limits object, specifying the upload ('limitUp')
              and download ('limitDown') speed in Kbps. These are only enforced if
              'settings' is set to 'custom'.
            suboptions:
              limitDown:
                description: The maximum download limit (integer, in Kbps).
                type: int
              limitUp:
                description: The maximum upload limit (integer, in Kbps).
                type: int
            type: dict
          settings:
            description: How bandwidth limits are applied by your rule. Can be one
              of 'network default', 'ignore' or 'custom'.
            type: str
        type: dict
    type: list
  trafficShapingEnabled:
    description: Whether traffic shaping rules are applied to clients on your SSID.
    type: bool
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for wireless updateNetworkWirelessSsidTrafficShapingRules
    description: Complete reference of the updateNetworkWirelessSsidTrafficShapingRules
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-network-wireless-ssid-traffic-shaping-rules
notes:
  - SDK Method used are
    wireless.Wireless.update_network_wireless_ssid_traffic_shaping_rules,
  - Paths used are
    put /networks/{networkId}/wireless/ssids/{number}/trafficShaping/rules,
"""

EXAMPLES = r"""
- name: Update all
  cisco.meraki.networks_wireless_ssids_traffic_shaping_rules:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    defaultRulesEnabled: true
    networkId: string
    number: string
    rules:
      - definitions:
          - type: host
            value: google.com
        dscpTagValue: 0
        pcpTagValue: 0
        perClientBandwidthLimits:
          bandwidthLimits:
            limitDown: 1000000
            limitUp: 1000000
          settings: custom
    trafficShapingEnabled: true
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "trafficShapingEnabled": true,
      "defaultRulesEnabled": true,
      "rules": [
        {
          "definitions": [
            {
              "type": "string",
              "value": "string"
            }
          ],
          "perClientBandwidthLimits": {
            "settings": "string",
            "bandwidthLimits": {
              "limitUp": 0,
              "limitDown": 0
            }
          },
          "dscpTagValue": 0,
          "pcpTagValue": 0
        }
      ]
    }
"""



#!/usr/bin/python
#
# Copyright (c) 2025 Klaas Demter (@Klaas-)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_monitorscheduledqueryrules
version_added: "3.11.0"
short_description: Create, update and delete Scheduled query rules
description:
    - Create, update and delete Scheduled query rules

options:
    name:
        description:
            - The name of the scheduled query rule you're creating/changing
        required: true
        type: str
    resource_group:
        description:
            - The name of the resource group
        required: true
        type: str
    location:
        description:
            - Location of the scheduled query rule
            - Scheduled query rules are usually reside at the location of your log analytics workspace.
        required: false
        type: str
    action:
        description:
            - The actions that will activate when the condition is met
        type: dict
        suboptions:
            azns_action:
                description: Azure action group reference
                type: dict
                suboptions:
                    action_group:
                        description:
                            - The list of the Action Group IDs
                        type: list
                        elements: str
                    email_subject:
                        description:
                            - Custom subject override for all email ids in Azure action group
                        type: str
            odata_type:
                description: Specifies the action
                type: str
                choices:
                    - AlertingAction
                    - LogToMetricAction
                    - >-
                      Microsoft.WindowsAzure.Management.Monitoring.Alerts.Models.Microsoft.AppInsights.Nexus.DataContracts.Resources.ScheduledQueryRules.AlertingAction
            severity:
                description: Severity of the alert
                type: int
                choices:
                    - 0
                    - 1
                    - 2
                    - 3
                    - 4
            trigger:
                description: The trigger condition that results in the alert rule being.
                type: dict
                suboptions:
                    metric_trigger:
                        description: The trigger condition that results in the alert rule being.
                        type: dict
                        suboptions:
                            metric_column:
                                description: Evaluation of metric on a particular column.
                                type: str
                            metric_trigger_type:
                                description: Metric Trigger Type.
                                type: str
                                choices:
                                    - Consecutive
                                    - Total
                            threshold:
                                description: The threshold of the metric trigger.
                                type: float
                            threshold_operator:
                                description: Evaluation operation for Metric.
                                type: str
                                choices:
                                    - GreaterThanOrEqual
                                    - GreaterThan
                                    - LessThan
                                    - Equal
                    threshold:
                        description: Result or count threshold based on which rule should be triggered.
                        type: float
                    threshold_operator:
                        description: Result or count threshold based on which rule should be triggered.
                        type: str
                        choices:
                            - GreaterThanOrEqual
                            - GreaterThan
                            - LessThan
    source:
        description: Data Source against which rule will Query Data.
        type: dict
        suboptions:
            authorized_resources:
                description: List of Resource referred into query.
                type: list
                elements: str
            data_source_id:
                description: The resource uri over which log search query is to be run.
                type: str
            query:
                description: Log search query. Required for action type - AlertingAction.
                type: str
            query_type:
                description: Set value to 'ResultCount'.
                type: str
                choices:
                    - ResultCount
    schedule:
        description: Schedule (Frequency, Time Window) for rule. Required for action type - AlertingAction.
        type: dict
        suboptions:
            frequency_in_minutes:
                description: Frequency (in minutes) at which rule condition should be evaluated.
                type: int
            time_window_in_minutes:
                description: Time window for which data needs to be fetched for query (should be greater than or equal to frequencyInMinutes).
                type: int
    enabled:
        description:
            - The flag which indicates whether the Log Search rule is enabled.
        type: str
        choices:
          - "False"
          - "True"
    auto_mitigate:
        description:
            - The flag that indicates whether the alert should be automatically resolved or not.
        type: bool
    description:
        description:
            - The description of the Log Search rule.
        type: str
    display_name:
        description:
            - The display name of the alert rule.
        type: str
    state:
        description:
            - State of the scheduled query rule
            - Use C(present) for creating/updating a scheduled query rule.
            - Use C(absent) for deleting a scheduled query rule.
        default: present
        type: str
        choices:
            - present
            - absent
extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags

author:
    - Klaas Demter (@Klaas-)
'''

EXAMPLES = '''
- name: Add a scheduled query rule
  azure.azcollection.azure_rm_monitorscheduledqueryrules:
    state: present
    resource_group: myrg
    name: scheduled_query_rule_name
    location: eastus
    action:
      azns_action:
        action_group:
          - /subscriptions/<sub-id>/resourceGroups/myrg/providers/microsoft.insights/actionGroups/action_group_name
        email_subject: ''
      odata_type: >-
        Microsoft.WindowsAzure.Management.Monitoring.Alerts.Models.Microsoft.AppInsights.Nexus.DataContracts.Resources.ScheduledQueryRules.AlertingAction
      severity: 2
      trigger:
        metric_trigger:
          metric_column: InstanceName,_ResourceId
          metric_trigger_type: Total
          threshold: 1
          threshold_operator: GreaterThanOrEqual
        threshold: 10
        threshold_operator: LessThan
    auto_mitigate: true
    description: 'Monitoring: Disks/logical volumes with less than 10% free space'
    display_name: Free_Space_Severity_2_Percentage
    enabled: "true"
    schedule:
      frequency_in_minutes: 5
      time_window_in_minutes: 5
    source:
      authorized_resources: []
      data_source_id: "/subscriptions/<sub-id>/resourceGroups/myrg/providers/Microsoft.OperationalInsights/workspaces/log_analytics_workspace_name"
      query: >
        Perf
        | where CounterName == "% Free Space"
        | where InstanceName != "total"
        | summarize AggregatedValue = max(CounterValue) by bin(TimeGenerated, 5m), _ResourceId, InstanceName
      query_type: ResultCount


- name: Add tag to existing scheduled query rule
  azure.azcollection.azure_rm_monitorscheduledqueryrules:
    state: present
    name: scheduled_query_rule_name
    resource_group: myrg
    append_tags: true
    tags:
      ThisIsAnAddedExampleTag: ExampleValue

- name: Delete a scheduled query rule
  azure.azcollection.azure_rm_monitorscheduledqueryrules:
    state: absent
    name: scheduled_query_rule_name
    resource_group: myrg
'''

RETURN = '''
scheduledqueryrule:
    description:
        - Details of the scheduled query rule
        - Is null on state==absent (scheduled query rule does not exist or will be deleted)
        - Assumes you make legal changes in check mode
    type: dict
    returned: always
    sample: {
        "action": {
            "azns_action": {
                "action_group": [
                    "/subscriptions/<sub-id>/resourceGroups/myrg/providers/microsoft.insights/actionGroups/action_group_name"
                ],
                "email_subject": ""
            },
            "odata_type": \
"Microsoft.WindowsAzure.Management.Monitoring.Alerts.Models.Microsoft.AppInsights.Nexus.DataContracts.Resources.ScheduledQueryRules.AlertingAction",
            "severity": "2",
            "trigger": {
                "metric_trigger": {
                    "metric_column": "InstanceName,_ResourceId",
                    "metric_trigger_type": "Total",
                    "threshold": 1.0,
                    "threshold_operator": "GreaterThanOrEqual"
                },
                "threshold": 10.0,
                "threshold_operator": "LessThan"
            }
        },
        "auto_mitigate": true,
        "created_with_api_version": "2018-04-16",
        "description": "Monitoring: Disks/logical volumes with less than 10% free space",
        "display_name": "scheduled_query_alert_name",
        "enabled": "true",
        "id": \
"/subscriptions/<sub-id>/resourceGroups/myrg/providers/microsoft.insights/scheduledqueryrules/scheduled_query_alert_name",
        "last_updated_time": "2025-10-30T16:07:30.773011Z",
        "location": "eastus",
        "name": "scheduled_query_alert_name",
        "provisioning_state": "Succeeded",
        "schedule": {
            "frequency_in_minutes": 5,
            "time_window_in_minutes": 5
        },
        "source": {
            "authorized_resources": [],
            "data_source_id": \
"/subscriptions/<sub-id>/resourceGroups/myrg/providers/Microsoft.OperationalInsights/workspaces/log_analytics_workspace_name",
            "query": "actual query",
            "query_type": "ResultCount"
        },
        "tags": {},
        "type": "Microsoft.Insights/scheduledQueryRules"
    }
'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ext import AzureRMModuleBaseExt

try:
    from azure.core.exceptions import ResourceNotFoundError

except ImportError:
    # This is handled in azure_rm_common
    pass

AZURE_OBJECT_CLASS = 'scheduledqueryrule'

azns_action_spec = dict(
    action_group=dict(type='list', elements='str'),
    email_subject=dict(type='str')
)

metric_trigger_spec = dict(
    metric_column=dict(type='str'),
    metric_trigger_type=dict(type='str', choices=['Consecutive', 'Total']),
    threshold=dict(type='float'),
    threshold_operator=dict(type='str', choices=[
        "GreaterThanOrEqual",
        "GreaterThan",
        "LessThan",
        "Equal"
    ])
)

trigger_spec = dict(
    metric_trigger=dict(type='dict', options=metric_trigger_spec),
    threshold=dict(type='float'),
    threshold_operator=dict(type='str', choices=[
        "GreaterThanOrEqual",
        "GreaterThan",
        "LessThan"
    ])
)

action_spec = dict(
    azns_action=dict(type='dict', options=azns_action_spec),
    odata_type=dict(type='str', choices=[
        'AlertingAction',
        'LogToMetricAction',
        'Microsoft.WindowsAzure.Management.Monitoring.Alerts.Models.Microsoft.AppInsights.Nexus.DataContracts.Resources.ScheduledQueryRules.AlertingAction'
    ]),
    severity=dict(type='int', choices=[0, 1, 2, 3, 4]),
    trigger=dict(type='dict', options=trigger_spec)
)

schedule_spec = dict(
    frequency_in_minutes=dict(type='int'),
    time_window_in_minutes=dict(type='int')
)

source_spec = dict(
    authorized_resources=dict(type='list', elements='str'),
    data_source_id=dict(type='str'),
    query=dict(type='str'),
    query_type=dict(type='str', choices=['ResultCount'])
)


class AzureRMMonitorscheduledqueryrule(AzureRMModuleBaseExt):
    """Information class for an Azure RM Scheduled query rules"""

    def __init__(self):
        self.module_arg_spec = dict(
            name=dict(type='str', required=True),
            resource_group=dict(type='str', required=True),
            location=dict(type='str'),
            action=dict(type='dict', options=action_spec),
            auto_mitigate=dict(type='bool'),
            description=dict(type='str'),
            display_name=dict(type='str'),
            enabled=dict(type='str', choices=['True', 'False']),
            schedule=dict(type='dict', options=schedule_spec),
            source=dict(type='dict', options=source_spec),
            state=dict(type='str', choices=['present', 'absent'], default='present')
        )

        self.name = None
        self.resource_group = None
        self.location = None
        self.tags = None
        self.action = None
        self.auto_mitigate = None
        self.description = None
        self.display_name = None
        self.enabled = None
        self.schedule = None
        self.source = None
        self.state = None
        self.log_path = None
        self.log_mode = None

        self.results = dict(
            changed=False,
            scheduledqueryrule=dict(),
            diff=dict(
                before=None,
                after=None
            )
        )

        super(AzureRMMonitorscheduledqueryrule, self).__init__(derived_arg_spec=self.module_arg_spec,
                                                               supports_check_mode=True,
                                                               supports_tags=True)

    def exec_module(self, **kwargs):
        """Main module execution method"""

        for key in list(self.module_arg_spec.keys()) + ['tags']:
            if hasattr(self, key):
                setattr(self, key, kwargs[key])

        # Defaults for variables
        result = None
        result_compare = dict(compare=[])
        before_dict = None

        # Get current scheduled query rule if it exists
        before_dict = self.get_scheduled_query_rule()

        # Create dict from input, without None values
        scheduled_query_rule_template = {
            "location": self.location,
            "action": self.action,
            "auto_mitigate": self.auto_mitigate,
            "description": self.description,
            "display_name": self.display_name,
            "enabled": self.enabled,
            "schedule": self.schedule,
            "source": self.source
        }
        # Filter out all None values
        scheduled_query_rule_input = {key: value for key, value in scheduled_query_rule_template.items() if value is not None}

        # Create/Update if state==present
        if self.state == 'present':
            if before_dict is None:
                # scheduled query rule does not exist, create
                # On creation default to location of resource group unless otherwise noted in input variables
                if not self.location:
                    resource_group = self.get_resource_group(self.resource_group)
                    scheduled_query_rule_input['location'] = resource_group.location
                # On creation input == what we send to api
                scheduled_query_rule_update = scheduled_query_rule_input
                # Needs to be extended by tags if set
                if self.tags:
                    scheduled_query_rule_update['tags'] = self.tags
                self.results['changed'] = True
                if self.check_mode:
                    # Check mode, skipping actual creation
                    pass
                else:
                    create_response = self.create_or_update(scheduled_query_rule_update)
            else:
                # scheduled query rule already exists, updating it
                # Dict for update is the union of existing object overwritten by input data
                scheduled_query_rule_update = before_dict | scheduled_query_rule_input

                # Enhanced with tags (special behaviour because of append_tags possibility)
                update_tags, update_tags_content = self.update_tags(before_dict.get('tags'))
                # Check if we need to update the scheduled query rule
                if update_tags or not self.default_compare({}, scheduled_query_rule_update, before_dict, '', result_compare):
                    scheduled_query_rule_update['tags'] = update_tags_content
                    # Need to create/update the scheduled query rule; changed -> True
                    self.results['changed'] = True
                    if self.check_mode:
                        # Check mode, skipping actual creation
                        pass
                    else:
                        create_response = self.create_or_update(scheduled_query_rule_update)

            if self.check_mode or not self.results['changed']:
                # When object was not updated or when running in check mode
                # assume scheduled_query_rule_update is resulting object
                result = scheduled_query_rule_update
            else:
                # otherwise take resulting new object from response of create call
                result = create_response

        # Delete scheduled query rule if state is absent and it exists
        # if it doesn't exist, it's already absent
        elif self.state == 'absent' and before_dict is not None:
            self.results['changed'] = True
            if self.check_mode:
                # do not delete in check mode
                pass
            else:
                self.delete()

        self.results['diff']['before'] = before_dict
        self.results['diff']['after'] = result
        self.results['scheduledqueryrule'] = result

        return self.results

    def get_scheduled_query_rule(self):
        '''
        Gets the properties of the specified scheduled query rule.

        :return: List of Scheduled query rules
        '''
        self.log("Checking if scheduled query rule {0} in resource group {1} is present".format(self.name,
                                                                                                self.resource_group))

        result = None
        response = None

        try:
            response = self.monitor_management_client_scheduled_query_rules.scheduled_query_rules.get(rule_name=self.name,
                                                                                                      resource_group_name=self.resource_group)
        except ResourceNotFoundError as ex:
            self.log("Could not find scheduled query rule {0} in resource group {1}".format(self.name, self.resource_group))
        if response:
            result = self.serialize_obj(response, AZURE_OBJECT_CLASS)

        return result

    def create_or_update(self, scheduled_query_rule_update):
        result = None
        response = None
        scheduled_query_rules = self.monitor_management_client_scheduled_query_rules.scheduled_query_rules
        try:
            response = scheduled_query_rules.create_or_update(resource_group_name=self.resource_group,
                                                              rule_name=self.name,
                                                              parameters=scheduled_query_rule_update,
                                                              logging_enable=False)
        except Exception as ex:
            self.fail("Error creating or update scheduled query rule {0} in resource group {1}: {2}".format(self.name, self.resource_group, str(ex)))

        if response:
            result = self.serialize_obj(response, AZURE_OBJECT_CLASS)

        return result

    def delete(self):
        response = None
        try:
            response = self.monitor_management_client_scheduled_query_rules.scheduled_query_rules.delete(resource_group_name=self.resource_group,
                                                                                                         rule_name=self.name)
        except Exception as ex:
            self.fail("Error deleting scheduled query rule {0} in resource group {1}: {2}".format(self.name, self.resource_group, str(ex)))

        return response


def main():
    """Main execution"""
    AzureRMMonitorscheduledqueryrule()


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# Copyright (c) 2025 xuzhang3 (@xuzhang3), Fred-sun (@Fred-sun)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_resourcehealthstates_info

version_added: "3.3.0"

short_description: Get or list the current availability status for the resource

description:
    - Get or list the current availability status for the resource.

options:
    resource_group:
        description:
            - Name of resource group.
        type: str
    resource_uri:
        description:
            - The fully qualified ID of the resource, including the resource name and resource type.
            - Currently the API support not nested and one nesting level resource types,
              such as '/subscriptions/{subscriptionId}/resourceGroups/{resource-group-name}/providers/{resource-provider-name}/
              {resource-type}/{resource-name}' and '/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/
              providers/{resource-provider-name}/{parentResourceType}/{parentResourceName}/{resourceType}/{resourceName}'.
        type: str
    list_all:
        description:
            - Whether lists all historical availability transitions and impacting events for a single resource.
        type: bool
        default: False

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - xuzhang3 (@xuzhang3)
    - Fred-sun (@Fred-sun)
'''

EXAMPLES = '''
- name: Lists the current availability status for all the resources in the subscription.
  azure_rm_resourcehealthstates_info:

- name: Lists the current availability status for all the resources in the resource group.
  azure_rm_resourcehealthstates_info:
    resource_group: "{{ resource_group }}"

- name: Gets current availability status for a single resource.
  azure_rm_resourcehealthstates_info:
    resource_uri: "/subscriptions/xxxx-xxx/resourcegroups/v-xisurg/providers/microsoft.compute/virtualmachines/testvm"

- name: Lists all historical availability transitions and impacting events for a single resource.
  azure_rm_resourcehealthstates_info:
    resource_uri: "/subscriptions/xxxx-xxx/resourcegroups/v-xisurg/providers/microsoft.compute/virtualmachines/testvm"
    list_all: true
'''

RETURN = '''
health_states:
    description:
        - The facts of availabilityStatus of a resource.
    type: complex
    returned: always
    contains:
        id:
            description:
                - Azure Resource Manager Identity for the availabilityStatuses resource.
            type: str
            returned: always
            sample: "/subscriptions/xxxx-xxx/resourcegroups/defaultresourcegroup-wus2/providers/microsoft.operationalinsights/workspaces/
                    defaultworkspace-e393adb3-b5be-4789-bdc9-848367f0d152-wus2/providers/Microsoft.ResourceHealth/availabilityStatuses/current"
        location:
            description:
                - Azure Resource Manager geo location of the resource.
            type: str
            returned: always
            sample: westus2
        type:
            description:
                - Microsoft.ResourceHealth/AvailabilityStatuses.
            type: str
            returned: always
            sample: "Microsoft.ResourceHealth/AvailabilityStatuses"
        properties:
            description:
                - Properties of availability state.
            type: complex
            contains:
                availability_state:
                    description:
                        - Availability status of the resource.
                    type: str
                    returned: always
                    sample: "Unknown"
                category:
                    description:
                        - When a context field is set to Platform, this field will reflect if the event was planned or unplanned.
                        - If the context field does not have a value of Platform, then this field will be ignored.
                    type: str
                    returned: always
                    sample: "Not Applicable"
                context:
                    description:
                        - This field is connected to the category field in this object.
                    type: str
                    returned: always
                    sample: "Not Applicable"
                occured_time:
                    description:
                        - Timestamp for when last change in health status occurred.
                    type: str
                    returned: always
                    sample: "2025-03-05T01:50:29.67632Z"
                reason_chronicity:
                    description:
                        - Chronicity of the availability transition.
                    type: str
                    returned: always
                    sample: "Transient"
                reason_type:
                    description:
                        - When the resource's availabilityState is Unavailable, it describes where the health impacting event was originated.
                    type: str
                    returned: always
                    sample: ""
                reported_time:
                    description:
                        - Timestamp for when the health was last checked.
                    type: str
                    returned: always
                    sample: "2025-03-05T01:50:29.67632Z"
                summary:
                    description:
                        - Summary description of the availability status.
                    type: str
                    returned: always
                    sample: "We're currently unable to determine the health of this workspace.
                            Either no queries were run on this workspace, or no data was ingested to this workspace recently."
                title:
                    description:
                        - Title description of the availability status.
                    type: str
                    returned: always
                    sample: Unkown
'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase


class AzureRMResourceHealthStatesInfo(AzureRMModuleBase):

    def __init__(self):

        self.module_args = dict(
            resource_group=dict(type='str'),
            resource_uri=dict(type='str'),
            list_all=dict(type='bool', default=False)
        )

        self.results = dict(
            changed=False,
            health_states=[]
        )

        self.resource_group = None
        self.resource_uri = None
        self.list_all = None
        required_if = [('list_all', True, ['resource_uri'])]

        super(AzureRMResourceHealthStatesInfo, self).__init__(
            derived_arg_spec=self.module_args,
            supports_check_mode=True,
            supports_tags=False,
            required_if=required_if,
            facts_module=True
        )

    def exec_module(self, **kwargs):
        for key in self.module_args:
            setattr(self, key, kwargs[key])

        if self.resource_uri is not None:
            if self.list_all:
                response = self.list_by_resource_uri(self.resource_uri)
            else:
                response = self.get_by_resource(self.resource_uri)
        elif self.resource_group is not None:
            response = self.list_by_resource_group(self.resource_group)
        else:
            response = self.list_by_subscription_id()

        self.results['health_states'] = [item.as_dict() for item in response]
        return self.results

    def get_by_resource(self, resource_uri):
        '''
        Gets current availability status for a single resource.
        :return: AvailabilityStatus or the result of cls(response)
        '''
        try:
            response = [self.resourcehealth_client.availability_statuses.get_by_resource(resource_uri)]
        except Exception as exc:
            self.fail('Error when gets current availability status for a single resource. Error msg: {0}'.format(exc.message or str(exc)))

        return response

    def list_by_resource_uri(self, resource_uri):
        '''
        Lists all historical availability transitions and impacting events for a single resource
        :return: An iterator like instance of either AvailabilityStatus or the result of cls(response)
        '''
        try:
            response = self.resourcehealth_client.availability_statuses.list(resource_uri)
        except Exception as exc:
            self.fail('Error when lists all historical availability transitions and impacting events for a single resource. Error msg: {0}'.format(exc))

        return response

    def list_by_resource_group(self, resource_group):
        '''
        Lists the current availability status for all the resources in the resource group.
        :return: An iterator like instance of either AvailabilityStatus or the result of cls(response)
        '''
        try:
            response = self.resourcehealth_client.availability_statuses.list_by_resource_group(resource_group)
        except Exception as exc:
            self.fail('Error when lists the current availability status for all the resources in the resource group. Error msg: {0}'.format(exc))

        return response

    def list_by_subscription_id(self):
        '''
        Lists the current availability status for all the resources in the subscription.
        :return: An iterator like instance of either AvailabilityStatus or the result of cls(response)
        '''
        try:
            response = self.resourcehealth_client.availability_statuses.list_by_subscription_id()
        except Exception as exc:
            self.fail('Error when lists the current availability status for all the resources in the subscription Error msg: {0}'.format(exc))

        return response


def main():
    """Main module execution code path"""

    AzureRMResourceHealthStatesInfo()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

"""
filter plugin file for ipaddr filters: ipv4
"""
from __future__ import absolute_import, division, print_function

from functools import partial

from ansible.errors import AnsibleFilterError

from ansible_collections.ansible.utils.plugins.module_utils.common.argspec_validate import (
    AnsibleArgSpecValidator,
)
from ansible_collections.ansible.utils.plugins.plugin_utils.base.ipaddr_utils import (
    _first_last,
    _need_netaddr,
    ipaddr,
)


__metaclass__ = type


try:
    from jinja2.filters import pass_environment
except ImportError:
    from jinja2.filters import environmentfilter as pass_environment

try:
    import netaddr

    HAS_NETADDR = True
except ImportError:
    # in this case, we'll make the filters return error messages (see bottom)
    HAS_NETADDR = False
else:

    class mac_linux(netaddr.mac_unix):
        pass

    mac_linux.word_fmt = "%.2x"

DOCUMENTATION = """
    name: next_nth_usable
    author: Ashwini Mhatre (@amhatre)
    version_added: "2.5.0"
    short_description: This filter returns the next nth usable ip within a network described by value.
    description:
        - This filter returns the next nth usable ip within a network described by value.
        - Use next_nth_usable to find the next nth usable IP address in relation to another within a range
    options:
        value:
            description:
            - subnets or individual address input for next_nth_usable plugin
            type: str
            required: True
        offset:
            description:
            - index value
            - next nth usable IP address
            type: int
    notes:
"""

EXAMPLES = r"""
#### examples
# Ipv4 filter plugin with different queries.
- name: next_nth_usable returns the second usable IP address for the given IP range
  debug:
    msg: "{{ '192.168.122.1/24' | ansible.utils.next_nth_usable(2) }}"

- name: If there is no usable address, it returns an empty string.
  debug:
    msg: "{{ '192.168.122.254/24' | ansible.utils.next_nth_usable(2) }}"

# TASK [next_nth_usable returns the second usable IP address for the given IP range] **************************
# task path: /Users/amhatre/ansible-collections/playbooks/test_next_nth_usable.yaml:9
# Loading collection ansible.utils from /Users/amhatre/ansible-collections/collections/ansible_collections/ansible/utils
# ok: [localhost] => {
#     "msg": "192.168.122.3"
# }
#
# TASK [If there is no usable address, it returns an empty string.] *******************************************
# task path: /Users/amhatre/ansible-collections/playbooks/test_next_nth_usable.yaml:14
# Loading collection ansible.utils from /Users/amhatre/ansible-collections/collections/ansible_collections/ansible/utils
# ok: [localhost] => {
#     "msg": ""
# }
"""

RETURN = """
  data:
    type: str
    description:
      - Returns the next nth usable ip within a network described by value.
"""


@pass_environment
def _next_nth_usable(*args, **kwargs):
    """This filter returns the next nth usable ip within a network described by value."""
    keys = ["value", "offset"]
    data = dict(zip(keys, args[1:]))
    data.update(kwargs)
    aav = AnsibleArgSpecValidator(data=data, schema=DOCUMENTATION, name="next_nth_usable")
    valid, errors, updated_data = aav.validate()
    if not valid:
        raise AnsibleFilterError(errors)
    return next_nth_usable(**updated_data)


def next_nth_usable(value, offset):
    """
    Returns the next nth usable ip within a network described by value.
    """
    try:
        v = None
        vtype = ipaddr(value, "type")
        if vtype == "address":
            v = ipaddr(value, "cidr")
        elif vtype == "network":
            v = ipaddr(value, "subnet")

        if v is not None:
            v = netaddr.IPNetwork(v)
        else:
            return False
    except Exception:
        return False

    if not isinstance(offset, int):
        raise AnsibleFilterError("Must pass in an integer")

    if v.size > 1:
        first_usable, last_usable = _first_last(v)
        nth_ip = int(netaddr.IPAddress(int(v.ip) + offset))
        if first_usable <= nth_ip <= last_usable:
            return str(netaddr.IPAddress(int(v.ip) + offset))


class FilterModule(object):
    """IP address and network manipulation filters"""

    filter_map = {
        # IP addresses and networks
        "next_nth_usable": _next_nth_usable,
    }

    def filters(self):
        """ipaddr filter"""
        if HAS_NETADDR:
            return self.filter_map
        else:
            return dict((f, partial(_need_netaddr, f)) for f in self.filter_map)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2020, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r'''
---
module: win_feature_info
version_added: '2.7.0'
short_description: Gather information about Windows features
description:
- Gather information about all or a specific installed Windows feature(s).
options:
  name:
    description:
    - If specified, this is used to match the C(name) of the Windows feature to get the info for.
    - Can be a wildcard to match multiple features but the wildcard will only be matched on the C(name) of the feature.
    - If omitted then all features will returned.
    type: str
    default: '*'
seealso:
- module: ansible.windows.win_feature
author:
- Larry Lane (@gamethis)
'''

EXAMPLES = r'''
- name: Get info for all installed features
  ansible.windows.win_feature_info:
  register: feature_info

- name: Get info for a single feature
  ansible.windows.win_feature_info:
    name: DNS
  register: feature_info

- name: Find all features that start with 'FS'
  ansible.windows.win_feature_info:
    name: FS*
'''

RETURN = r'''
exists:
  description: Whether any features were found based on the criteria specified.
  returned: always
  type: bool
  sample: true
features:
  description:
  - A list of feature(s) that were found based on the criteria.
  - Will be an empty list if no features were found.
  returned: always
  type: list
  elements: dict
  contains:
    name:
      description:
      - Name of feature found.
      type: str
      sample: AD-Certificate
    display_name:
      description:
      - The Display name of feature found.
      type: str
      sample: Active Directory Certificate Services
    description:
      description:
      - The description of the feature.
      type: str
      sample: Example description of the Windows feature.
    installed:
      description:
      - Whether the feature by C(name) is installed.
      type: bool
      sample: false
    install_state:
      description:
      - The Install State of C(name).
      - Values will be one of C(Available), C(Removed), C(Installed).
      type: str
      sample: Installed
    feature_type:
      description:
      - The Feature Type of C(name).
      - Values will be one of C(Role), C(Role Service), C(Feature).
      type: str
      sample: Feature
    path:
      description:
      - The Path of C(name) feature.
      type: str
      sample: WoW64 Support
    depth:
      description:
      - Depth of C(name) feature.
      type: int
      sample: 1
    depends_on:
      description:
      - The command line that will be run when a C(run_command) failure action is fired.
      type: list
      elements: str
      sample: ['Web-Static-Content', 'Web-Default-Doc']
    parent:
      description:
      - The parent of feature C(name) if present.
      type: str
      sample: PowerShellRoot
    server_component_descriptor:
      description:
      - Descriptor of C(name) feature.
      type: str
      sample: ServerComponent_AD_Certificate
    sub_features:
      description:
      - List of sub features names of feature C(name).
      type: list
      elements: str
      sample: ['WAS-Process-Model', 'WAS-NET-Environment', 'WAS-Config-APIs']
    system_service:
      description:
      - The name of the service installed by feature C(name).
      type: list
      elements: str
      sample: ['iisadmin', 'w3svc']
    best_practices_model_id:
      description:
      - BestPracticesModelId for feature C(name).
      type: str
      sample: Microsoft/Windows/UpdateServices
    event_query:
      description:
      - The EventQuery for feature C(name).
      - This will be C(null) if None Present
      type: str
      sample: IPAMServer.Events.xml
    post_configuration_needed:
      description:
      - Tells if Post Configuration is needed for feature C(name).
      type: bool
      sample: False
    additional_info:
      description:
      - A list of privileges that the feature requires and will run with
      type: dict
      contains:
        major_version:
          description:
          - Major Version of feature C(name).
          type: int
          sample: 8
        minor_version:
          description:
          - Minor Version of feature C(name).
          type: int
          sample: 0
        number_id_version:
          description:
          - Numberic Id of feature C(name).
          type: int
          sample: 16
        install_name:
          description:
          - The action to perform once triggered, can be C(start_feature) or C(stop_feature).
          type: str
          sample: ADCertificateServicesRole
'''



# Copyright (c) 2021 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import typing as t

from ansible_collections.community.dns.plugins.module_utils._zone_record_api import (
    DNSAPIError,
)

if t.TYPE_CHECKING:
    from collections.abc import Sequence  # pragma: no cover

    from ._argspec import OptionProvider  # pragma: no cover
    from ._provider import ProviderInformation  # pragma: no cover
    from ._record import DNSRecord, IDNSRecord, RecordIDT  # pragma: no cover
    from ._zone import ZoneIDT  # pragma: no cover
    from ._zone_record_api import ZoneRecordAPI  # pragma: no cover

    class _Result(t.TypedDict, t.Generic[RecordIDT]):  # pragma: no cover
        deleted: list[DNSRecord[RecordIDT]]  # pragma: no cover
        changed: list[DNSRecord[RecordIDT]]  # pragma: no cover
        created: list[DNSRecord[RecordIDT]]  # pragma: no cover


def bulk_apply_changes(
    api: ZoneRecordAPI[ZoneIDT, RecordIDT],
    provider_information: ProviderInformation,
    options: OptionProvider,
    zone_id: ZoneIDT,
    records_to_delete: Sequence[DNSRecord[RecordIDT]] | None = None,
    records_to_change: Sequence[DNSRecord[RecordIDT]] | None = None,
    records_to_create: Sequence[IDNSRecord[RecordIDT | None]] | None = None,
    stop_early_on_errors: bool = True,
) -> tuple[bool, list[DNSAPIError], _Result[RecordIDT]]:
    """
    Update multiple records. If an operation failed, raise a DNSAPIException.

    @param api: A ZoneRecordAPI instance
    @param provider_information: A ProviderInformation object.
    @param options: A object compatible with ModuleOptionProvider that gives access to the module/plugin
                    options.
    @param zone_id: Zone ID to apply changes to
    @param records_to_delete: Optional list of DNS records to delete (DNSRecord)
    @param records_to_change: Optional list of DNS records to change (DNSRecord)
    @param records_to_create: Optional list of DNS records to create (DNSRecord)
    @param bulk_threshold: Minimum number of changes for using the bulk API instead of the regular API
    @param stop_early_on_errors: If set to ``True``, try to stop changes after the first error happens.
                                 This might only work on some APIs.
    @return A tuple (changed, errors, success) where ``changed`` is a boolean which indicates whether a
            change was made, ``errors`` is a list of ``DNSAPIError`` instances for the errors occurred,
            and ``success`` is a dictionary with three lists ``success['deleted']``,
            ``success['changed']`` and ``success['created']``, which list all records that were deleted,
            changed and created, respectively.
    """
    records_to_delete = records_to_delete or []
    records_to_change = records_to_change or []
    records_to_create = records_to_create or []

    has_change = False
    errors: list[DNSAPIError] = []

    bulk_threshold = 2
    if provider_information.supports_bulk_actions():
        bulk_threshold = options.get_option("bulk_operation_threshold")

    success: _Result[RecordIDT] = {
        "deleted": [],
        "changed": [],
        "created": [],
    }

    # Delete records
    if len(records_to_delete) >= bulk_threshold:
        results = api.delete_records(
            {zone_id: records_to_delete}, stop_early_on_errors=stop_early_on_errors
        )
        result = results.get(zone_id) or []
        for record, deleted, failed in result:
            has_change |= deleted
            if failed is not None:
                errors.append(failed)
            if deleted:
                success["deleted"].append(record)
        if errors and stop_early_on_errors:
            return has_change, errors, success
    else:
        for record in records_to_delete:
            try:
                deleted = api.delete_record(zone_id, record)
                has_change |= deleted
                if deleted:
                    success["deleted"].append(record)
            except DNSAPIError as e:
                errors.append(e)
                if stop_early_on_errors:
                    return has_change, errors, success

    # Change records
    if len(records_to_change) >= bulk_threshold:
        results = api.update_records(
            {zone_id: records_to_change}, stop_early_on_errors=stop_early_on_errors
        )
        result = results.get(zone_id) or []
        for record, changed, failed in result:
            has_change |= changed
            if failed is not None:
                errors.append(failed)
            if changed:
                success["changed"].append(record)
        if errors and stop_early_on_errors:
            return has_change, errors, success
    else:
        for record in records_to_change:
            try:
                record = api.update_record(zone_id, record)
                has_change = True
                success["changed"].append(record)
            except DNSAPIError as e:
                errors.append(e)
                if stop_early_on_errors:
                    return has_change, errors, success

    # Create records
    if len(records_to_create) >= bulk_threshold:
        create_results = api.add_records(
            {zone_id: records_to_create}, stop_early_on_errors=stop_early_on_errors
        )
        create_result = create_results.get(zone_id) or []
        for crecord, created, failed in create_result:
            has_change |= created
            if failed is not None:
                errors.append(failed)
            if created:
                success["created"].append(crecord)  # type: ignore  # if created is True, crecord is a DNSRecord
        if errors and stop_early_on_errors:
            return has_change, errors, success
    else:
        for create_record in records_to_create:
            try:
                created_record = api.add_record(zone_id, create_record)
                has_change = True
                success["created"].append(created_record)
            except DNSAPIError as e:
                errors.append(e)
                if stop_early_on_errors:
                    return has_change, errors, success

    return has_change, errors, success



#!/usr/bin/python

# Copyright (c) 2020, FERREIRA Christophe <christophe.ferreira@cnaf.fr>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: nomad_job
author: FERREIRA Christophe (@chris93111)
version_added: "1.3.0"
short_description: Launch a Nomad Job
description:
  - Launch a Nomad job.
  - Stop a Nomad job.
  - Force start a Nomad job.
requirements:
  - python-nomad
extends_documentation_fragment:
  - community.general._nomad
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  name:
    description:
      - Name of job for delete, stop and start job without source.
      - Name of job for delete, stop and start job without source.
      - Either this or O(content) must be specified.
    type: str
  state:
    description:
      - Deploy or remove job.
    choices: ["present", "absent"]
    required: true
    type: str
  force_start:
    description:
      - Force job to started.
    type: bool
    default: false
  content:
    description:
      - Content of Nomad job.
      - Either this or O(name) must be specified.
    type: str
  content_format:
    description:
      - Type of content of Nomad job.
    choices: ["hcl", "json"]
    default: hcl
    type: str
seealso:
  - name: Nomad jobs documentation
    description: Complete documentation for Nomad API jobs.
    link: https://www.nomadproject.io/api-docs/jobs/
"""

EXAMPLES = r"""
- name: Create job
  community.general.nomad_job:
    host: localhost
    state: present
    content: "{{ lookup('ansible.builtin.file', 'job.hcl') }}"
    timeout: 120

- name: Connect with port to create job
  community.general.nomad_job:
    host: localhost
    port: 4645
    state: present
    content: "{{ lookup('ansible.builtin.file', 'job.hcl') }}"
    timeout: 120

- name: Stop job
  community.general.nomad_job:
    host: localhost
    state: absent
    name: api

- name: Force job to start
  community.general.nomad_job:
    host: localhost
    state: present
    name: api
    timeout: 120
    force_start: true
"""

import json

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.text.converters import to_native

from ansible_collections.community.general.plugins.module_utils._nomad import argument_spec as nomad_argument_spec
from ansible_collections.community.general.plugins.module_utils._nomad import nomad, setup_nomad_client


def run():
    module = AnsibleModule(
        argument_spec={
            **nomad_argument_spec,
            "state": dict(required=True, choices=["present", "absent"]),
            "name": dict(type="str"),
            "content_format": dict(choices=["hcl", "json"], default="hcl"),
            "content": dict(type="str"),
            "force_start": dict(type="bool", default=False),
        },
        supports_check_mode=True,
        mutually_exclusive=[["name", "content"]],
        required_one_of=[["name", "content"]],
    )

    nomad_client = setup_nomad_client(module)

    if module.params["state"] == "present":
        if module.params["name"] and not module.params["force_start"]:
            module.fail_json(msg="For start job with name, force_start is needed")

        changed = False
        if module.params["content"]:
            if module.params["content_format"] == "json":
                job_json = module.params["content"]
                try:
                    job_json = json.loads(job_json)
                except ValueError as e:
                    module.fail_json(msg=f"{e}")
                job = {"job": job_json}
                try:
                    job_id = job_json.get("ID")
                    if job_id is None:
                        module.fail_json(msg="Cannot retrieve job with ID None")
                    plan = nomad_client.job.plan_job(job_id, job, diff=True)
                    if plan["Diff"].get("Type") != "None":
                        changed = True
                        if not module.check_mode:
                            result = nomad_client.jobs.register_job(job)
                        else:
                            result = plan
                    else:
                        result = plan
                except Exception as e:
                    module.fail_json(msg=f"{e}")

            if module.params["content_format"] == "hcl":
                try:
                    job_hcl = module.params["content"]
                    job_json = nomad_client.jobs.parse(job_hcl)
                    job = {"job": job_json}
                except nomad.api.exceptions.BadRequestNomadException as err:
                    module.fail_json(msg=f"{err.nomad_resp.reason} {err.nomad_resp.text}")
                try:
                    job_id = job_json.get("ID")
                    plan = nomad_client.job.plan_job(job_id, job, diff=True)
                    if plan["Diff"].get("Type") != "None":
                        changed = True
                        if not module.check_mode:
                            result = nomad_client.jobs.register_job(job)
                        else:
                            result = plan
                    else:
                        result = plan
                except Exception as e:
                    module.fail_json(msg=f"{e}")

        if module.params["force_start"]:
            try:
                if module.params["name"]:
                    job_name = module.params["name"]
                else:
                    job_name = job_json["Name"]
                job_json = nomad_client.job.get_job(job_name)
                if job_json["Status"] == "running":
                    result = job_json
                else:
                    job_json["Status"] = "running"
                    job_json["Stop"] = False
                    job = {"job": job_json}
                    if not module.check_mode:
                        result = nomad_client.jobs.register_job(job)
                    else:
                        result = nomad_client.validate.validate_job(job)
                        if result.status_code != 200:
                            module.fail_json(msg=to_native(result.text))
                        result = json.loads(result.text)
                    changed = True
            except Exception as e:
                module.fail_json(msg=f"{e}")

    if module.params["state"] == "absent":
        try:
            if module.params["name"] is not None:
                job_name = module.params["name"]
            else:
                if module.params["content_format"] == "hcl":
                    job_json = nomad_client.jobs.parse(module.params["content"])
                    job_name = job_json["Name"]
                if module.params["content_format"] == "json":
                    job_json = module.params["content"]
                    job_name = job_json["Name"]
            job = nomad_client.job.get_job(job_name)
            if job["Status"] == "dead":
                changed = False
                result = job
            else:
                if not module.check_mode:
                    result = nomad_client.job.deregister_job(job_name)
                else:
                    result = job
                changed = True
        except Exception as e:
            module.fail_json(msg=f"{e}")

    module.exit_json(changed=changed, result=result)


def main():
    run()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_appliance_uplinks_usage_by_network_info
short_description: Information module for organizations _appliance _uplinks _usage
  _by _network
description:
  - Get all organizations _appliance _uplinks _usage _by _network. - > Get the sent
    and received bytes for each uplink of all MX and Z networks within an organization.
    If more than one device was active during the specified timespan, then the sent
    and received bytes will be aggregated by interface.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module_info
author: Francisco Munoz (@fmunoz)
options:
  headers:
    description: Additional headers.
    type: dict
  organizationId:
    description:
      - OrganizationId path parameter. Organization ID.
    type: str
  t0:
    description:
      - T0 query parameter. The beginning of the timespan for the data. The maximum
        lookback period is 30 days from today.
    type: str
  t1:
    description:
      - T1 query parameter. The end of the timespan for the data. T1 can be a maximum
        of 14 days after t0.
    type: str
  timespan:
    description:
      - >
        Timespan query parameter. The timespan for which the information will be fetched.
        If specifying timespan, do not specify parameters t0 and t1. The value must
        be in seconds and be less than or equal to 14 days. The default is 1 day.
    type: float
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for appliance getOrganizationApplianceUplinksUsageByNetwork
    description: Complete reference of the getOrganizationApplianceUplinksUsageByNetwork
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!get-organization-appliance-uplinks-usage-by-network
notes:
  - SDK Method used are
    appliance.Appliance.get_organization_appliance_uplinks_usage_by_network,
  - Paths used are
    get /organizations/{organizationId}/appliance/uplinks/usage/byNetwork,
"""

EXAMPLES = r"""
- name: Get all organizations _appliance _uplinks _usage _by _network
  cisco.meraki.organizations_appliance_uplinks_usage_by_network_info:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    t0: string
    t1: string
    timespan: 0
    organizationId: string
  register: result
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: list
  elements: dict
  sample: >
    [
      {
        "networkId": "string",
        "name": "string",
        "byUplink": [
          {
            "serial": "string",
            "interface": "string",
            "sent": 0,
            "received": 0
          }
        ]
      }
    ]
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: cloudwatchlogs_log_group_info
version_added: 5.0.0
short_description: Get information about log_group in CloudWatchLogs
description:
  - Lists the specified log groups. You can list all your log groups or filter the results by prefix.
  - This module was originally added to C(community.aws) in release 1.0.0.
author:
  - Willian Ricardo (@willricardo) <willricardo@gmail.com>
options:
  log_group_name:
    description:
      - The name or prefix of the log group to filter by.
    type: str
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
# Note: These examples do not set authentication details, see the AWS Guide for details.
- amazon.aws.cloudwatchlogs_log_group_info:
    log_group_name: test-log-group
"""

RETURN = r"""
log_groups:
    description: Return the list of complex objects representing log groups.
    returned: success
    type: complex
    contains:
        log_group_name:
            description: The name of the log group.
            returned: always
            type: str
        creation_time:
            description: The creation time of the log group.
            returned: always
            type: int
        retention_in_days:
            description: The number of days to retain the log events in the specified log group.
            returned: always
            type: int
        metric_filter_count:
            description: The number of metric filters.
            returned: always
            type: int
        arn:
            description: The Amazon Resource Name (ARN) of the log group.
            returned: always
            type: str
        stored_bytes:
            description: The number of bytes stored.
            returned: always
            type: str
        kms_key_id:
            description: The Amazon Resource Name (ARN) of the CMK to use when encrypting log data.
            returned: always
            type: str
        tags:
            description: A dictionary representing the tags on the log group.
            returned: always
            type: dict
            version_added: 4.0.0
            version_added_collection: community.aws
"""

try:
    import botocore
except ImportError:
    pass  # Handled by AnsibleAWSModule

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.retries import AWSRetry


@AWSRetry.exponential_backoff()
def list_tags_log_group_with_backoff(client, log_group_name):
    return client.list_tags_log_group(logGroupName=log_group_name)


@AWSRetry.exponential_backoff()
def describe_log_groups_with_backoff(client, **kwargs):
    paginator = client.get_paginator("describe_log_groups")
    return paginator.paginate(**kwargs).build_full_result()


def describe_log_group(client, log_group_name, module):
    params = {}
    if log_group_name:
        params["logGroupNamePrefix"] = log_group_name
    try:
        desc_log_group = describe_log_groups_with_backoff(client, **params)
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg=f"Unable to describe log group {log_group_name}")

    for log_group in desc_log_group["logGroups"]:
        log_group_name = log_group["logGroupName"]
        try:
            tags = list_tags_log_group_with_backoff(client, log_group_name)
        except is_boto3_error_code("AccessDeniedException"):
            tags = {}
            module.warn(f"Permission denied listing tags for log group {log_group_name}")
        except (
            botocore.exceptions.ClientError,
            botocore.exceptions.BotoCoreError,
        ) as e:  # pylint: disable=duplicate-except
            module.fail_json_aws(e, msg=f"Unable to describe tags for log group {log_group_name}")
        log_group["tags"] = tags.get("tags", {})

    return desc_log_group


def main():
    argument_spec = dict(
        log_group_name=dict(),
    )

    module = AnsibleAWSModule(argument_spec=argument_spec, supports_check_mode=True)

    try:
        logs = module.client("logs")
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Failed to connect to AWS")

    desc_log_group = describe_log_group(client=logs, log_group_name=module.params["log_group_name"], module=module)
    final_log_group_snake = []

    for log_group in desc_log_group["logGroups"]:
        final_log_group_snake.append(camel_dict_to_snake_dict(log_group, ignore_list=["tags"]))

    desc_log_group_result = dict(changed=False, log_groups=final_log_group_snake)
    module.exit_json(**desc_log_group_result)


if __name__ == "__main__":
    main()



#
# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The nxos_snmp_server config file.
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to its desired end-state is
created.
"""

from copy import deepcopy

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.resource_module import (
    ResourceModule,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    dict_merge,
    get_from_dict,
)

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.facts.facts import Facts
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.rm_templates.snmp_server import (
    Snmp_serverTemplate,
)


class Snmp_server(ResourceModule):
    """
    The nxos_snmp_server config class
    """

    def __init__(self, module):
        super(Snmp_server, self).__init__(
            empty_fact_val={},
            facts_module=Facts(module),
            module=module,
            resource="snmp_server",
            tmplt=Snmp_serverTemplate(),
        )
        self.parsers = [
            "aaa_user.cache_timeout",
            "contact",
            "context",
            "counter.enable",
            "counter.cache.timeout",
            "drop.unknown_engine_id",
            "drop.unknown_user",
            "traps.aaa",
            "traps.bgp",
            "traps.bridge.newroot",
            "traps.bridge.topologychange",
            "traps.callhome.event_notify",
            "traps.callhome.smtp_send_fail",
            "traps.cfs.merge_failure",
            "traps.cfs.state_change_notif",
            "traps.config.ccmCLIRunningConfigChanged",
            "traps.entity.cefcMIBEnableStatusNotification",
            "traps.entity.entity_fan_status_change",
            "traps.entity.entity_mib_change",
            "traps.entity.entity_module_inserted",
            "traps.entity.entity_module_status_change",
            "traps.entity.entity_power_out_change",
            "traps.entity.entity_power_status_change",
            "traps.entity.entity_sensor",
            "traps.entity.entity_unrecognised_module",
            "traps.feature_control.featureOpStatusChange",
            "traps.feature_control.ciscoFeatOpStatusChange",
            "traps.generic.coldStart",
            "traps.generic.warmStart",
            "traps.license.notify_license_expiry",
            "traps.license.notify_license_expiry_warning",
            "traps.license.notify_licensefile_missing",
            "traps.license.notify_no_license_for_feature",
            "traps.link.cErrDisableInterfaceEventRev1",
            "traps.link.cieLinkDown",
            "traps.link.cieLinkUp",
            "traps.link.cisco_xcvr_mon_status_chg",
            "traps.link.cmn_mac_move_notification",
            "traps.link.delayed_link_state_change",
            "traps.link.extended_linkDown",
            "traps.link.extended_linkUp",
            "traps.link.linkDown",
            "traps.link.linkUp",
            "traps.mmode.cseMaintModeChangeNotify",
            "traps.mmode.cseNormalModeChangeNotify",
            "traps.ospf",
            "traps.ospfv3",
            "traps.rf.redundancy_framework",
            "traps.rmon.fallingAlarm",
            "traps.rmon.hcFallingAlarm",
            "traps.rmon.hcRisingAlarm",
            "traps.rmon.risingAlarm",
            "traps.snmp.authentication",
            "traps.storm_control.cpscEventRev1",
            "traps.storm_control.trap_rate",
            "traps.stpx.inconsistency",
            "traps.stpx.root_inconsistency",
            "traps.stpx.loop_inconsistency",
            "traps.syslog.message_generated",
            "traps.sysmgr.cseFailSwCoreNotifyExtended",
            "traps.system.clock_change_notification",
            "traps.upgrade.upgradeJobStatusNotify",
            "traps.upgrade.upgradeOpNotifyOnCompletion",
            "traps.vtp.notifs",
            "traps.vtp.vlancreate",
            "traps.vtp.vlandelete",
            "engine_id.local",
            "global_enforce_priv",
            "location",
            "mib.community_map",
            "packetsize",
            "protocol.enable",
            "source_interface.informs",
            "source_interface.traps",
            "system_shutdown",
            "tcp_session",
        ]

    def execute_module(self):
        """Execute the module

        :rtype: A dictionary
        :returns: The result from module execution
        """
        if self.state not in ["parsed", "gathered"]:
            self.generate_commands()
            self.run_commands()
        return self.result

    def generate_commands(self):
        """Generate configuration commands to send based on
        want, have and desired state.
        """
        wantd = self._list_to_dict(self.want)
        haved = self._list_to_dict(self.have)

        # if state is merged, merge want onto have and then compare
        if self.state == "merged":
            wantd = dict_merge(haved, wantd)

        # this ensures that if user sets `enable: True` for a trap
        # all suboptions for that trap are set to True
        for x in [
            "traps.aaa",
            "traps.bridge",
            "traps.callhome",
            "traps.cfs",
            "traps.config",
            "traps.entity",
            "traps.feature_control",
            "traps.generic",
            "traps.license",
            "traps.link",
            "traps.mmode",
            "traps.rf",
            "traps.rmon",
            "traps.snmp",
            "traps.storm_control",
            "traps.stpx",
            "traps.syslog",
            "traps.sysmgr",
            "traps.system",
            "traps.upgrade",
            "traps.vtp",
        ]:
            entry = get_from_dict(wantd, x)
            if entry and entry.get("enable", False):
                key = x.split(".")
                wantd[key[0]][key[1]].pop("enable")
                for i in self.parsers:
                    if i.startswith(x):
                        key = i.split(".")
                        wantd[key[0]][key[1]][key[2]] = True

        # if state is deleted, empty out wantd and set haved to wantd
        if self.state == "deleted":
            wantd = {}

        self._compare(want=wantd, have=haved)

    def _compare(self, want, have):
        """Leverages the base class `compare()` method and
        populates the list of commands to be run by comparing
        the `want` and `have` data with the `parsers` defined
        for the Snmp_server network resource.
        """
        self.compare(parsers=self.parsers, want=want, have=have)
        self._compare_lists(want=want, have=have)

    def _compare_lists(self, want, have):
        """
        Compare list of dictionaries
        """
        for x in ["users.auth", "users.use_acls", "hosts", "communities"]:
            wantx = get_from_dict(want, x) or {}
            havex = get_from_dict(have, x) or {}
            for wkey, wentry in wantx.items():
                hentry = havex.pop(wkey, {})
                if wentry != hentry:
                    self.addcmd(wentry, x)
            # remove superfluous items
            for _k, hv in havex.items():
                self.addcmd(hv, x, negate=True)

    def _list_to_dict(self, data):
        def _build_key(x):
            key = set()
            for k, v in x.items():
                if isinstance(v, dict):
                    for sk, sv in v.items():
                        if isinstance(sv, dict):
                            for ssk, ssv in sv.items():
                                key.add(sk + "_" + ssk + "_" + str(ssv))
                        else:
                            key.add(sk + "_" + str(sv))
                else:
                    key.add(k + "_" + str(v))
            return tuple(sorted(key))

        tmp = deepcopy(data)
        if "communities" in tmp:
            tmp["communities"] = {_build_key(entry): entry for entry in tmp["communities"]}
        if "users" in tmp:
            if "auth" in tmp["users"]:
                tmp["users"]["auth"] = {_build_key(entry): entry for entry in tmp["users"]["auth"]}
            if "use_acls" in tmp["users"]:
                tmp["users"]["use_acls"] = {
                    entry["user"]: entry for entry in tmp["users"]["use_acls"]
                }
        if "hosts" in tmp:
            tmp["hosts"] = {_build_key(entry): entry for entry in tmp["hosts"]}
        return tmp



#
# -*- coding: utf-8 -*-
# Copyright 2024 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The iosxr_vrf_address_family config file.
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to its desired end-state is
created.
"""

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.resource_module import (
    ResourceModule,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    dict_merge,
)

from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.facts.facts import Facts
from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.rm_templates.vrf_address_family import (
    Vrf_address_familyTemplate,
)


class Vrf_address_family(ResourceModule):
    """
    The iosxr_vrf_address_family config class
    """

    def __init__(self, module):
        super(Vrf_address_family, self).__init__(
            empty_fact_val={},
            facts_module=Facts(module),
            module=module,
            resource="vrf_address_family",
            tmplt=Vrf_address_familyTemplate(),
        )
        self.parsers = [
            "address_family",
            "export.route_policy",
            "export.route_target",
            "export.to.default_vrf.route_policy",
            "export.to.vrf.allow_imported_vpn",
            "import_config.route_target",
            "import_config.route_policy",
            "import_config.from_config.bridge_domain.advertise_as_vpn",
            "import_config.from_config.default_vrf.route_policy",
            "import_config.from_config.vrf.advertise_as_vpn",
            "maximum.prefix",
        ]

    def execute_module(self):
        """Execute the module

        :rtype: A dictionary
        :returns: The result from module execution
        """
        if self.state not in ["parsed", "gathered"]:
            self.generate_commands()
            self.run_commands()
        return self.result

    def generate_commands(self):
        """Generate configuration commands to send based on
        want, have and desired state.
        """
        wantd = self.want
        haved = self.have

        wantd = self._vrf_list_to_dict(wantd)
        haved = self._vrf_list_to_dict(haved)

        # if state is merged, merge want into have and then compare
        if self.state == "merged":
            wantd = dict_merge(haved, wantd)

        # if state is deleted, empty out wantd and set haved to wantd
        if self.state == "deleted":
            for vrfk, vrfv in haved.items():
                for afk, afv in vrfv.get("address_families", {}).items():
                    adrf = wantd.get(vrfk, {}).get("address_families", {})
                    if afk in adrf or not adrf:
                        self.addcmd(
                            {"name": vrfk},
                            "name",
                            False,
                        )
                        self.addcmd(
                            {"afi": afv.get("afi"), "safi": afv.get("safi")},
                            "address_family",
                            True,
                        )

        if self.state in ["overridden"]:
            for vrfk, vrfv in haved.items():
                for k, have in (vrfv.get("address_families", {}) or {}).items():
                    wantx = wantd.get(vrfk, {}).get("address_families", {})
                    if k not in wantx:
                        self.addcmd(
                            {"name": vrfk},
                            "name",
                            False,
                        )
                        self.addcmd(
                            {"afi": have.get("afi"), "safi": have.get("safi")},
                            "address_family",
                            False,
                        )
                        self.compare(parsers=self.parsers, want={}, have=have)

        if self.state != "deleted":
            self._compare(want=wantd, have=haved)

    def _compare(self, want, have):
        """Leverages the base class `compare()` method and
        populates the list of commands to be run by comparing
        the `want` and `have` data with the `parsers` defined
        for the Vrf network resource.
        """
        for name, entry in want.items():
            begin = len(self.commands)
            vrf_want = entry
            vrf_have = have.pop(name, {})
            self._compare_afs(vrf_want, vrf_have)
            if len(self.commands) != begin:
                self.commands.insert(begin, "vrf {0}".format(name))

    def _compare_afs(self, want, have):
        """Custom handling of afs option
        :params want: the want VRF dictionary
        :params have: the have VRF dictionary
        """
        waafs = want.get("address_families", {})
        haafs = have.get("address_families", {})
        for afk, afv in waafs.items():
            begin = len(self.commands)
            self._compare_single_af(want=afv, have=haafs.get(afk, {}))
            if len(self.commands) != begin:
                self.commands.insert(begin, f"address-family {afv.get('afi')} {afv.get('safi')}")

    def _compare_single_af(self, want, have):
        """Custom handling of single af option
        :params want: the want VRF dictionary
        :params have: the have VRF dictionary
        """
        self.compare(parsers=self.parsers[1:], want=want, have=have)

    def _vrf_list_to_dict(self, entry):
        """Convert list of items to dict of items
        for efficient diff calculation.
        :params entry: data dictionary
        """

        for vrf in entry:
            if "address_families" in vrf:
                vrf["address_families"] = {
                    f"{x['afi']}_{x.get('safi')}": x for x in vrf["address_families"]
                }
        entry = {x["name"]: x for x in entry}
        return entry



#!/usr/bin/python
# Copyright (c) 2024 Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = """
module: acme_certificate_order_create
author: Felix Fontein (@felixfontein)
version_added: 2.24.0
short_description: Create an ACME v2 order
description:
  - Creates an ACME v2 order. This is the first step of obtaining a new certificate
    with the L(ACME protocol,https://tools.ietf.org/html/rfc8555) from a Certificate
    Authority such as L(Let's Encrypt,https://letsencrypt.org/).
    This module does not support ACME v1, the original version of the ACME protocol
    before standardization.
  - The current implementation supports the V(http-01), V(dns-01), V(dns-account-01), V(dns-persist-01),
    and V(tls-alpn-01) challenges.
  - This module needs to be used in conjunction with the
    M(community.crypto.acme_certificate_order_validate) and.
    M(community.crypto.acme_certificate_order_finalize) module.
    An order can be effectively deactivated with the
    M(community.crypto.acme_certificate_deactivate_authz) module.
    Note that both modules require the output RV(order_uri) of this module.
  - To create or modify ACME accounts, use the M(community.crypto.acme_account) module.
    This module will I(not) create or update ACME accounts.
  - Between the call of this module and M(community.crypto.acme_certificate_order_finalize),
    you have to fulfill the required steps for the chosen challenge by whatever means necessary.
    For V(http-01) that means creating the necessary challenge file on the destination webserver.
    For V(dns-01), V(dns-account-01), and V(dns-persist-01) the necessary DNS records have to be created.
    For V(tls-alpn-01) the necessary certificate has to be created and served.
    It is I(not) the responsibility of this module to perform these steps.
  - For details on how to fulfill these challenges, you might have to read through
    L(the main ACME specification,https://tools.ietf.org/html/rfc8555#section-8)
    and the L(TLS-ALPN-01 specification,https://www.rfc-editor.org/rfc/rfc8737.html#section-3).
    Also, consider the examples provided for this module.
  - The module includes support for IP identifiers according to
    L(RFC 8738,https://www.rfc-editor.org/rfc/rfc8738.html) ACME extension.
  - The module B(experimentally) supports the V(dns-account-01) challenge type according to
    L(acme-dns-account-label draft 02, https://datatracker.ietf.org/doc/html/draft-ietf-acme-dns-account-label-02).
    Note that the supported draft version can change at any time,
    and changes will only be considered breaking once the draft reached RFC status.
  - The module B(experimentally) supports the V(dns-persist-01) challenge type according to
    L(acme-dns-persist draft 01, https://www.ietf.org/archive/id/draft-ietf-acme-dns-persist-01.html).
    Note that the supported draft version can change at any time,
    and changes will only be considered breaking once the draft reached RFC status.
seealso:
  - module: community.crypto.acme_certificate_order_validate
    description: Validate pending authorizations of an ACME order.
  - module: community.crypto.acme_certificate_order_finalize
    description: Finalize an ACME order after satisfying the challenges.
  - module: community.crypto.acme_certificate_order_info
    description: Obtain information for an ACME order.
  - module: community.crypto.acme_certificate_deactivate_authz
    description: Deactivate all authorizations (authz) of an ACME order, effectively deactivating
                 the order itself.
  - module: community.crypto.acme_certificate_renewal_info
    description: Determine whether a certificate should be renewed.
  - name: The Let's Encrypt documentation
    description: Documentation for the Let's Encrypt Certification Authority.
                 Provides useful information for example on rate limits.
    link: https://letsencrypt.org/docs/
  - name: Automatic Certificate Management Environment (ACME)
    description: The specification of the ACME protocol (RFC 8555).
    link: https://tools.ietf.org/html/rfc8555
  - name: ACME TLS ALPN Challenge Extension
    description: The specification of the V(tls-alpn-01) challenge (RFC 8737).
    link: https://www.rfc-editor.org/rfc/rfc8737.html
  - module: community.crypto.acme_challenge_cert_helper
    description: Helps preparing V(tls-alpn-01) challenges.
  - module: community.crypto.openssl_privatekey
    description: Can be used to create private keys (both for certificates and accounts).
  - module: community.crypto.openssl_privatekey_pipe
    description: Can be used to create private keys without writing it to disk (both for certificates and accounts).
  - module: community.crypto.openssl_csr
    description: Can be used to create a Certificate Signing Request (CSR).
  - module: community.crypto.openssl_csr_pipe
    description: Can be used to create a Certificate Signing Request (CSR) without writing it to disk.
  - module: community.crypto.acme_account
    description: Allows to create, modify or delete an ACME account.
  - module: community.crypto.acme_inspect
    description: Allows to debug problems.
extends_documentation_fragment:
  - community.crypto._acme.basic
  - community.crypto._acme.account
  - community.crypto._acme.certificate
  - community.crypto._attributes
  - community.crypto._attributes.actiongroup_acme
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
  idempotent:
    support: none
options:
  deactivate_authzs:
    description:
      - "Deactivate authentication objects (authz) when issuing the certificate
         failed."
      - "Authentication objects are bound to an account key and remain valid
         for a certain amount of time, and can be used to issue certificates
         without having to re-authenticate the domain. This can be a security
         concern."
    type: bool
    default: true
  replaces_cert_id:
    description:
      - If provided, will request the order to replace the certificate identified by this certificate ID
        according to L(Section 5 of RFC 9773, https://www.rfc-editor.org/rfc/rfc9773.html#section-5).
      - This certificate ID must be computed as specified in
        L(Section 4.1 of RFC 9773, https://www.rfc-editor.org/rfc/rfc9773.html#section-4.1).
        It is returned as return value RV(community.crypto.acme_certificate_renewal_info#module:cert_id) of the
        M(community.crypto.acme_certificate_renewal_info) module.
      - ACME servers might refuse to create new orders that indicate to replace a certificate for which
        an active replacement order already exists. This can happen if this module is used to create an order,
        and then the playbook/role fails in case the challenges cannot be set up. If the playbook/role does not
        record the order data to continue with the existing order, but tries to create a new one on the next run,
        creating the new order might fail. If O(order_creation_error_strategy=fail) this will make the module fail.
        O(order_creation_error_strategy=auto) and O(order_creation_error_strategy=retry_without_replaces_cert_id)
        will avoid this by leaving away C(replaces) on retries.
      - If O(order_creation_error_strategy=fail), for the above reason, this option should only be used
        if the role/playbook using it keeps track of order data accross restarts, or if it takes care to
        deactivate orders whose processing is aborted. Orders can be deactivated with the
        M(community.crypto.acme_certificate_deactivate_authz) module.
    type: str
  profile:
    description:
      - Chose a specific profile for certificate selection. The available profiles depend on the CA.
      - See L(a blog post by Let's Encrypt, https://letsencrypt.org/2025/01/09/acme-profiles/) and
        L(draft-aaron-acme-profiles-01, https://datatracker.ietf.org/doc/draft-aaron-acme-profiles/)
        for more information.
    type: str
  order_creation_error_strategy:
    description:
      - Selects the error handling strategy for ACME protocol errors if creating a new ACME order fails.
    type: str
    choices:
      auto:
        - An unspecified algorithm that tries to be clever.
        - Right now identical to V(retry_without_replaces_cert_id).
      always:
        - Always retry, until the limit in O(order_creation_max_retries) has been reached.
      fail:
        - Simply fail in case of errors. Do not attempt to retry.
      retry_without_replaces_cert_id:
        - If O(replaces_cert_id) is present, creating the order will be tried again without C(replaces).
        - The only exception is an error of type C(urn:ietf:params:acme:error:alreadyReplaced), that indicates that
          the certificate was already replaced. This usually means something went wrong and the user should investigate.
    default: auto
  order_creation_max_retries:
    description:
      - Depending on the strategy selected in O(order_creation_error_strategy), will retry creating new orders
        for at most the specified amount of times.
    type: int
    default: 3
"""

EXAMPLES = r"""
---
### Example with HTTP-01 challenge ###

- name: Create a challenge for sample.com using a account key from a variable
  community.crypto.acme_certificate_order_create:
    account_key_content: "{{ account_private_key }}"
    csr: /etc/pki/cert/csr/sample.com.csr
  register: sample_com_challenge

# Alternative first step:
- name: Create a challenge for sample.com using a account key from Hashi Vault
  community.crypto.acme_certificate_order_create:
    account_key_content: >-
      {{ lookup('community.hashi_vault.hashi_vault', 'secret=secret/account_private_key:value') }}
    csr: /etc/pki/cert/csr/sample.com.csr
  register: sample_com_challenge

# Alternative first step:
- name: Create a challenge for sample.com using a account key file
  community.crypto.acme_certificate_order_create:
    account_key_src: /etc/pki/cert/private/account.key
    csr_content: "{{ lookup('file', '/etc/pki/cert/csr/sample.com.csr') }}"
  register: sample_com_challenge

# Perform the necessary steps to fulfill the challenge. For example:
#
# - name: Copy http-01 challenges
#   ansible.builtin.copy:
#     dest: /var/www/{{ item.identifier }}/{{ item.challenges['http-01'].resource }}
#     content: "{{ item.challenges['http-01'].resource_value }}"
#   loop: "{{ sample_com_challenge.challenge_data }}"
#   when: "'http-01' in item.challenges"

- name: Let the challenge be validated
  community.crypto.acme_certificate_order_validate:
    account_key_src: /etc/pki/cert/private/account.key
    order_uri: "{{ sample_com_challenge.order_uri }}"
    challenge: http-01

- name: Retrieve the cert and intermediate certificate
  community.crypto.acme_certificate_order_finalize:
    account_key_src: /etc/pki/cert/private/account.key
    csr: /etc/pki/cert/csr/sample.com.csr
    order_uri: "{{ sample_com_challenge.order_uri }}"
    cert_dest: /etc/httpd/ssl/sample.com.crt
    fullchain_dest: /etc/httpd/ssl/sample.com-fullchain.crt
    chain_dest: /etc/httpd/ssl/sample.com-intermediate.crt

---
### Example with DNS challenge against production ACME server ###

- name: Create a challenge for sample.com using a account key file.
  community.crypto.acme_certificate_order_create:
    acme_directory: https://acme-v01.api.letsencrypt.org/directory
    account_key_src: /etc/pki/cert/private/account.key
    csr: /etc/pki/cert/csr/sample.com.csr
  register: sample_com_challenge

# Perform the necessary steps to fulfill the challenge. For example:
#
# - name: Create DNS records for dns-01 challenges
#   community.aws.route53:
#     zone: sample.com
#     record: "{{ item.key }}"
#     type: TXT
#     ttl: 60
#     state: present
#     wait: true
#     # Note: item.value is a list of TXT entries, and route53
#     # requires every entry to be enclosed in quotes
#     value: "{{ item.value | map('community.dns.quote_txt', always_quote=true) | list }}"
#   loop: "{{ sample_com_challenge.challenge_data_dns | dict2items }}"

- name: Let the challenge be validated
  community.crypto.acme_certificate_order_validate:
    acme_directory: https://acme-v01.api.letsencrypt.org/directory
    account_key_src: /etc/pki/cert/private/account.key
    order_uri: "{{ sample_com_challenge.order_uri }}"
    challenge: dns-01

- name: Retrieve the cert and intermediate certificate
  community.crypto.acme_certificate_order_finalize:
    acme_directory: https://acme-v01.api.letsencrypt.org/directory
    account_key_src: /etc/pki/cert/private/account.key
    csr: /etc/pki/cert/csr/sample.com.csr
    order_uri: "{{ sample_com_challenge.order_uri }}"
    cert_dest: /etc/httpd/ssl/sample.com.crt
    fullchain_dest: /etc/httpd/ssl/sample.com-fullchain.crt
    chain_dest: /etc/httpd/ssl/sample.com-intermediate.crt
"""

RETURN = """
challenge_data:
  description:
    - For every identifier, provides the challenge information.
    - Only challenges which are not yet valid are returned.
  returned: changed
  type: list
  elements: dict
  contains:
    identifier:
      description:
        - The identifier for this challenge.
      type: str
      sample: example.com
    identifier_type:
      description:
        - The identifier's type.
        - V(dns) for DNS names, and V(ip) for IP addresses.
      type: str
      choices:
        - dns
        - ip
      sample: dns
    challenges:
      description:
        - Information for different challenge types supported for this identifier.
        - Note that the keys are not valid Jinja2 identifiers.
      type: dict
      contains:
        http-01:
          description:
            - Information for V(http-01) authorization.
            - The server needs to make the path RV(challenge_data[].challenges.http-01.resource)
              accessible via HTTP (which might redirect to HTTPS). A C(GET) operation to this path
              needs to provide the value from RV(challenge_data[].challenges.http-01.resource_value).
          returned: if the identifier supports V(http-01) authorization
          type: dict
          contains:
            resource:
              description:
                - The path the value has to be provided under.
              returned: success
              type: str
              sample: .well-known/acme-challenge/evaGxfADs6pSRb2LAv9IZf17Dt3juxGJ-PCt92wr-oA
            resource_value:
              description:
                - The value the resource has to produce for the validation.
              returned: success
              type: str
              sample: IlirfxKKXA...17Dt3juxGJ-PCt92wr-oA
        dns-01:
          description:
            - Information for V(dns-01) authorization.
            - A DNS TXT record needs to be created with the record name RV(challenge_data[].challenges.dns-01.record)
              and value RV(challenge_data[].challenges.dns-01.resource_value).
          returned: if the identifier supports V(dns-01) authorization
          type: dict
          contains:
            resource:
              description:
                - Always contains the string V(_acme-challenge).
              returned: success
              type: str
              sample: _acme-challenge
            resource_value:
              description:
                - The value the resource has to produce for the validation.
              returned: success
              type: str
              sample: IlirfxKKXA...17Dt3juxGJ-PCt92wr-oA
            record:
              description: The full DNS record's name for the challenge.
              returned: success
              type: str
              sample: _acme-challenge.example.com
        dns-account-01:
          description:
            - Information for V(dns-account-01) authorization.
            - A DNS TXT record needs to be created with the record name RV(challenge_data[].challenges.dns-01.record)
              and value RV(challenge_data[].challenges.dns-01.resource_value).
            - B(Note) that this return value is B(experimental) and can change in a breaking way also in minor releases,
              until the acme-dns-account-label draft is in a more stable state.
          returned: if the identifier supports V(dns-account-01) authorization
          version_added: 3.2.0
          type: dict
          contains:
            resource:
              description:
                - Always ends with the string V(._acme-challenge).
              returned: success
              type: str
              sample: _ujmmovf2vn55tgye._acme-challenge
            resource_value:
              description:
                - The value the resource has to produce for the validation.
              returned: success
              type: str
              sample: IlirfxKKXA...17Dt3juxGJ-PCt92wr-oA
            record:
              description: The full DNS record's name for the challenge.
              returned: success
              type: str
              sample: _ujmmovf2vn55tgye._acme-challenge.example.com
        dns-persist-01:
          description:
            - Information for V(dns-persist-01) authorization.
            - A DNS TXT record needs to be created with the record name V(_validation-persist.<domain>).
              See the P(community.crypto.acme_dns_persist_record#filter) for how to create the record's content.
            - B(Note) that this return value is B(experimental) and can change in a breaking way also in minor releases,
              until the acme-dns-persist draft is in a more stable state.
          returned: if the identifier supports V(dns-persist-01) authorization
          version_added: 3.2.0
          type: dict
          contains:
            account_uri:
              description:
                - The account URI that must be mentioned in the DNS TXT record.
                - B(Note) that the type of this return value might change in a breaking way also in minor releases,
                  until the acme-dns-persist draft is in a more stable state.
                  This challenge field is currently under discussion in the ACME WG.
              returned: success
              type: str
              sample: https://ca.example/acct/123
            issuer_domain_names:
              description:
                - One of the issuer domain names must be mentioned in the DNS TXT record.
              returned: success
              type: list
              elements: str
              sample:
                - letsencrypt.org
        tls-alpn-01:
          description:
            - Information for V(tls-alpn-01) authorization.
            - A certificate needs to be created for the DNS name RV(challenge_data[].challenges.tls-alpn-01.resource)
              with acmeValidation X.509 extension of value RV(challenge_data[].challenges.tls-alpn-01.resource_value).
              This certificate needs to be served when the application-layer protocol C(acme-tls/1) is negotiated for
              a HTTPS connection to port 443 with the SNI extension for the domain name
              (RV(challenge_data[].challenges.tls-alpn-01.resource_original)) being validated.
            - See U(https://www.rfc-editor.org/rfc/rfc8737.html#section-3) for details.
          returned: if the identifier supports V(tls-alpn-01) authorization
          type: dict
          contains:
            resource:
              description:
                - The DNS name for DNS identifiers, and the reverse DNS mapping (RFC1034, RFC3596) for IP addresses.
              returned: success
              type: str
              sample: example.com
            resource_original:
              description:
                - The original identifier including type identifier.
              returned: success
              type: str
              sample: dns:example.com
            resource_value:
              description:
                - The value the resource has to produce for the validation.
                - "B(Note:) this return value contains a Base64 encoded version of the correct
                   binary blob which has to be put into the acmeValidation X.509 extension; see
                   U(https://www.rfc-editor.org/rfc/rfc8737.html#section-3) for details. To do this,
                   you might need the P(ansible.builtin.b64decode#filter) Jinja filter to extract
                   the binary blob from this return value."
              returned: success
              type: str
              sample: AAb=
challenge_data_dns:
  description:
    - List of TXT values per DNS record for V(dns-01) challenges.
    - Only challenges which are not yet valid are returned.
  returned: success
  type: dict
challenge_data_dns_account:
  description:
    - List of TXT values per DNS record for V(dns-account-01) challenges.
    - Only challenges which are not yet valid are returned.
  returned: success
  type: dict
  version_added: 3.2.0
order_uri:
  description: ACME order URI.
  returned: success
  type: str
account_uri:
  description: ACME account URI.
  returned: success
  type: str
"""

import typing as t

from ansible_collections.community.crypto.plugins.module_utils._acme.acme import (
    create_backend,
    create_default_argspec,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.certificate import (
    ACMECertificateClient,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.errors import (
    ModuleFailException,
)


def main() -> t.NoReturn:
    argument_spec = create_default_argspec(with_certificate=True)
    argument_spec.update_argspec(
        deactivate_authzs={"type": "bool", "default": True},
        replaces_cert_id={"type": "str"},
        profile={"type": "str"},
        order_creation_error_strategy={
            "type": "str",
            "default": "auto",
            "choices": ["auto", "always", "fail", "retry_without_replaces_cert_id"],
        },
        order_creation_max_retries={"type": "int", "default": 3},
    )
    module = argument_spec.create_ansible_module()

    backend = create_backend(module, needs_acme_v2=False)

    try:
        client = ACMECertificateClient(module=module, backend=backend)

        profile = module.params["profile"]
        if profile is not None:
            meta_profiles = (client.client.directory.get("meta") or {}).get(
                "profiles"
            ) or {}
            if not meta_profiles:
                raise ModuleFailException(
                    msg='The ACME CA does not support profiles. Please omit the "profile" option.'
                )
            if profile not in meta_profiles:
                raise ModuleFailException(
                    msg=f"The ACME CA does not support selected profile {profile!r}."
                )

        order = None
        done = False
        try:
            order = client.create_order(
                replaces_cert_id=module.params["replaces_cert_id"], profile=profile
            )
            client.check_that_authorizations_can_be_used(order)
            done = True
        finally:
            if module.params["deactivate_authzs"] and order and not done:
                client.deactivate_authzs(order)
        data, data_dns, data_dns_account = client.get_challenges_data(order)
        module.exit_json(
            changed=True,
            order_uri=order.url,
            account_uri=client.client.account_uri,
            challenge_data=data,
            challenge_data_dns=data_dns,
            challenge_data_dns_account=data_dns_account,
        )
    except ModuleFailException as e:
        e.do_fail(module=module)


if __name__ == "__main__":
    main()
